from __future__ import annotations

import logging
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Literal,
    Mapping,
    Optional,
    Type,
    overload,
)

import great_expectations.exceptions as gx_exceptions
from great_expectations._docs_decorators import public_api
from great_expectations.exceptions import (
    GXCloudConfigurationError,
)
from great_expectations.exceptions.exceptions import DataContextRequiredError

logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    # needed until numpy min version 1.20
    from typing_extensions import TypeAlias

    from great_expectations.alias_types import PathStr
    from great_expectations.core.config_provider import _ConfigurationProvider
    from great_expectations.core.factory import ValidationDefinitionFactory
    from great_expectations.core.factory.suite_factory import SuiteFactory
    from great_expectations.data_context import (
        AbstractDataContext,
        CloudDataContext,
        EphemeralDataContext,
        FileDataContext,
    )
    from great_expectations.data_context.store import (
        ExpectationsStore,
        ValidationResultsStore,
    )
    from great_expectations.data_context.store.checkpoint_store import CheckpointStore
    from great_expectations.data_context.store.validation_definition_store import (
        ValidationDefinitionStore,
    )
    from great_expectations.data_context.types.base import DataContextConfig
    from great_expectations.datasource.datasource_dict import DatasourceDict
    from great_expectations.datasource.fluent.batch_request import BatchRequest
    from great_expectations.validator.validator import Validator

ContextModes: TypeAlias = Literal["file", "cloud", "ephemeral"]


class ProjectManager:
    """Singleton class to manage projects in the global namespace."""

    __project: AbstractDataContext | None

    def __init__(self):
        self.__project = None

    def get_project(  # noqa: PLR0913 # FIXME CoP
        self,
        project_config: DataContextConfig | Mapping | None = None,
        context_root_dir: PathStr | None = None,
        project_root_dir: PathStr | None = None,
        runtime_environment: dict | None = None,
        cloud_base_url: str | None = None,
        cloud_access_token: str | None = None,
        cloud_organization_id: str | None = None,
        cloud_mode: bool | None = None,
        user_agent_str: str | None = None,
        mode: ContextModes | None = None,
    ) -> AbstractDataContext:
        self.__project = self._build_context(
            project_config=project_config,
            context_root_dir=context_root_dir,
            project_root_dir=project_root_dir,
            runtime_environment=runtime_environment,
            cloud_base_url=cloud_base_url,
            cloud_access_token=cloud_access_token,
            cloud_organization_id=cloud_organization_id,
            cloud_mode=cloud_mode,
            user_agent_str=user_agent_str,
            mode=mode,
        )
        return self.__project

    def set_project(self, project: AbstractDataContext | None) -> None:
        self.__project = project

    @property
    def _project(self) -> AbstractDataContext:
        if not self.__project:
            raise DataContextRequiredError()
        return self.__project

    def get_expectations_store(self) -> ExpectationsStore:
        return self._project.expectations_store

    def get_checkpoints_store(self) -> CheckpointStore:
        return self._project.checkpoint_store

    def get_validation_results_store(self) -> ValidationResultsStore:
        return self._project.validation_results_store

    def get_validation_definition_store(self) -> ValidationDefinitionStore:
        return self._project.validation_definition_store

    def get_validation_definitions_factory(self) -> ValidationDefinitionFactory:
        return self._project.validation_definitions

    def get_datasources(self) -> DatasourceDict:
        return self._project.data_sources.all()

    def get_validator(self, batch_request: BatchRequest) -> Validator:
        return self._project.get_validator(batch_request=batch_request)

    def get_suite_factory(self) -> SuiteFactory:
        return self._project.suites

    def is_using_cloud(self) -> bool:
        from great_expectations.data_context import CloudDataContext

        return isinstance(self._project, CloudDataContext)

    def build_data_docs(
        self,
        site_names: list[str] | None = None,
        resource_identifiers: list | None = None,
        dry_run: bool = False,
        build_index: bool = True,
    ) -> dict:
        return self._project.build_data_docs(
            site_names=site_names,
            resource_identifiers=resource_identifiers,
            dry_run=dry_run,
            build_index=build_index,
        )

    def get_docs_sites_urls(
        self,
        resource_identifier: Any | None = None,
        site_name: str | None = None,
        only_if_exists: bool = True,
        site_names: list[str] | None = None,
    ) -> list[dict[str, Optional[str]]]:
        return self._project.get_docs_sites_urls(
            resource_identifier=resource_identifier,
            site_name=site_name,
            only_if_exists=only_if_exists,
            site_names=site_names,
        )

    def get_config_provider(self) -> _ConfigurationProvider:
        return self._project.config_provider

    def _build_context(  # noqa: PLR0913 # FIXME CoP
        self,
        project_config: DataContextConfig | Mapping | None = None,
        context_root_dir: PathStr | None = None,
        project_root_dir: PathStr | None = None,
        runtime_environment: dict | None = None,
        cloud_base_url: str | None = None,
        cloud_access_token: str | None = None,
        cloud_organization_id: str | None = None,
        cloud_mode: bool | None = None,
        user_agent_str: str | None = None,
        mode: ContextModes | None = None,
    ) -> AbstractDataContext:
        project_config = self._prepare_project_config(project_config)

        if mode == "file" and not context_root_dir and not project_root_dir:
            project_root_dir = Path.cwd()

        param_lookup: dict[ContextModes | None, dict] = {
            "ephemeral": dict(
                project_config=project_config,
                runtime_environment=runtime_environment,
                user_agent_str=user_agent_str,
            ),
            "file": dict(
                project_config=project_config,
                context_root_dir=context_root_dir,
                project_root_dir=project_root_dir,
                runtime_environment=runtime_environment,
                user_agent_str=user_agent_str,
            ),
            "cloud": dict(
                project_config=project_config,
                context_root_dir=context_root_dir,
                project_root_dir=project_root_dir,
                runtime_environment=runtime_environment,
                cloud_base_url=cloud_base_url,
                cloud_access_token=cloud_access_token,
                cloud_organization_id=cloud_organization_id,
                user_agent_str=user_agent_str,
                cloud_mode=True,
            ),
            None: dict(
                project_config=project_config,
                context_root_dir=context_root_dir,
                project_root_dir=project_root_dir,
                runtime_environment=runtime_environment,
                cloud_base_url=cloud_base_url,
                cloud_access_token=cloud_access_token,
                cloud_organization_id=cloud_organization_id,
                user_agent_str=user_agent_str,
                cloud_mode=cloud_mode,
                mode=mode,
            ),
        }
        try:
            kwargs = param_lookup[mode]
        except KeyError:
            raise ValueError(f"Unknown mode {mode}. Please choose one of: ephemeral, file, cloud.")  # noqa: TRY003 # FIXME CoP

        from great_expectations.data_context.data_context import (
            AbstractDataContext,
            CloudDataContext,
            EphemeralDataContext,
            FileDataContext,
        )

        expected_ctx_types: dict[
            ContextModes | None,
            Type[CloudDataContext]
            | Type[EphemeralDataContext]
            | Type[FileDataContext]
            | Type[AbstractDataContext],
        ] = {
            "ephemeral": EphemeralDataContext,
            "file": FileDataContext,
            "cloud": CloudDataContext,
            None: AbstractDataContext,  # type: ignore[type-abstract] # FIXME CoP
        }

        context_fn_map: dict[ContextModes | None, Callable] = {
            "ephemeral": self._get_ephemeral_context,
            "file": self._get_file_context,
            "cloud": self._get_cloud_context,
            None: self._get_default_context,
        }

        context_fn = context_fn_map[mode]
        context = context_fn(**kwargs)

        expected_type = expected_ctx_types[mode]
        if not isinstance(context, expected_type):
            # example I want an ephemeral context but the presence of a GX_CLOUD env var gives me a cloud context  # noqa: E501 # FIXME CoP
            raise ValueError(  # noqa: TRY003, TRY004 # FIXME CoP
                f"Provided mode {mode} returned context of type {type(context).__name__} instead of {expected_type.__name__}; please check your input arguments."  # noqa: E501 # FIXME CoP
            )

        return context

    def _get_default_context(  # noqa: PLR0913 # FIXME CoP
        self,
        project_config: DataContextConfig | None = None,
        context_root_dir: PathStr | None = None,
        project_root_dir: PathStr | None = None,
        runtime_environment: dict | None = None,
        cloud_base_url: str | None = None,
        cloud_access_token: str | None = None,
        cloud_organization_id: str | None = None,
        user_agent_str: str | None = None,
        cloud_mode: bool | None = None,
        mode: ContextModes | None = None,
    ) -> AbstractDataContext:
        """Infer which type of DataContext a user wants based on available parameters."""
        # First, check for GX Cloud conditions
        cloud_context = self._get_cloud_context(
            project_config=project_config,
            context_root_dir=context_root_dir,
            project_root_dir=project_root_dir,
            runtime_environment=runtime_environment,
            cloud_mode=cloud_mode,
            cloud_base_url=cloud_base_url,
            cloud_access_token=cloud_access_token,
            cloud_organization_id=cloud_organization_id,
            user_agent_str=user_agent_str,
            mode=mode,
        )

        if cloud_context:
            return cloud_context

        # Second, check for a context_root_dir to determine if using a filesystem
        file_context = self._get_file_context(
            project_config=project_config,
            context_root_dir=context_root_dir,
            project_root_dir=project_root_dir,
            runtime_environment=runtime_environment,
            user_agent_str=user_agent_str,
        )
        if file_context:
            return file_context

        # Finally, default to ephemeral
        return self._get_ephemeral_context(
            project_config=project_config,
            runtime_environment=runtime_environment,
            user_agent_str=user_agent_str,
        )

    def _prepare_project_config(
        self,
        project_config: DataContextConfig | Mapping | None,
    ) -> DataContextConfig | None:
        from great_expectations.data_context.data_context import AbstractDataContext
        from great_expectations.data_context.types.base import DataContextConfig

        # If available and applicable, convert project_config mapping into a rich config type
        if project_config:
            project_config = AbstractDataContext.get_or_create_data_context_config(project_config)
        assert project_config is None or isinstance(project_config, DataContextConfig), (
            "project_config must be of type Optional[DataContextConfig]"
        )

        return project_config

    def _get_cloud_context(  # noqa: PLR0913 # FIXME CoP
        self,
        project_config: DataContextConfig | Mapping | None = None,
        context_root_dir: PathStr | None = None,
        project_root_dir: PathStr | None = None,
        runtime_environment: dict | None = None,
        cloud_base_url: str | None = None,
        cloud_access_token: str | None = None,
        cloud_organization_id: str | None = None,
        user_agent_str: str | None = None,
        cloud_mode: bool | None = None,
        mode: ContextModes | None = None,
    ) -> CloudDataContext | None:
        if cloud_mode is False:
            return None  # user has specifically disabled cloud mode, so we don't check for env vars

        from great_expectations.data_context.data_context import CloudDataContext

        config_available = CloudDataContext.is_cloud_config_available(
            cloud_base_url=cloud_base_url,
            cloud_access_token=cloud_access_token,
            cloud_organization_id=cloud_organization_id,
        )

        if config_available:
            return CloudDataContext(
                project_config=project_config,
                runtime_environment=runtime_environment,
                context_root_dir=context_root_dir,
                project_root_dir=project_root_dir,
                cloud_base_url=cloud_base_url,
                cloud_access_token=cloud_access_token,
                cloud_organization_id=cloud_organization_id,
                user_agent_str=user_agent_str,
            )
        elif (
            mode != "cloud" and not cloud_mode
        ):  # cloud mode not specified, and env vars not available
            return None
        else:
            raise GXCloudConfigurationError(  # noqa: TRY003 # one time exception
                "Unable to create a CloudDataContext due to one or more missing environment "
                "variables: GX_CLOUD_ORGANIZATION_ID, GX_CLOUD_ACCESS_TOKEN"
            )

    def _get_file_context(
        self,
        project_config: DataContextConfig | None = None,
        context_root_dir: PathStr | None = None,
        project_root_dir: PathStr | None = None,
        runtime_environment: dict | None = None,
        user_agent_str: str | None = None,
    ) -> FileDataContext | None:
        from great_expectations.data_context.data_context import FileDataContext

        try:
            return FileDataContext(
                project_config=project_config,
                context_root_dir=context_root_dir,
                project_root_dir=project_root_dir,
                runtime_environment=runtime_environment,
                user_agent_str=user_agent_str,
            )
        except gx_exceptions.ConfigNotFoundError:
            logger.info("Could not find local file-backed GX project")
            return None

    def _get_ephemeral_context(
        self,
        project_config: DataContextConfig | None = None,
        runtime_environment: dict | None = None,
        user_agent_str: str | None = None,
    ) -> EphemeralDataContext:
        from great_expectations.data_context.data_context import EphemeralDataContext
        from great_expectations.data_context.types.base import (
            DataContextConfig,
            InMemoryStoreBackendDefaults,
        )

        if not project_config:
            project_config = DataContextConfig(
                store_backend_defaults=InMemoryStoreBackendDefaults(init_temp_docs_sites=True)
            )

        return EphemeralDataContext(
            project_config=project_config,
            runtime_environment=runtime_environment,
            user_agent_str=user_agent_str,
        )


# global singleton
project_manager = ProjectManager()


@overload
def get_context(
    project_config: DataContextConfig | Mapping | None = ...,
    context_root_dir: None = ...,
    project_root_dir: None = ...,
    runtime_environment: dict | None = ...,
    cloud_base_url: None = ...,
    cloud_access_token: None = ...,
    cloud_organization_id: None = ...,
    cloud_mode: Literal[False] | None = ...,
    user_agent_str: str | None = ...,
    mode: Literal["ephemeral"] = ...,
) -> EphemeralDataContext: ...


@overload
def get_context(
    project_config: DataContextConfig | Mapping | None = ...,
    context_root_dir: PathStr = ...,  # If context_root_dir is provided, project_root_dir shouldn't be  # noqa: E501 # FIXME CoP
    project_root_dir: None = ...,
    runtime_environment: dict | None = ...,
    cloud_base_url: None = ...,
    cloud_access_token: None = ...,
    cloud_organization_id: None = ...,
    cloud_mode: Literal[False] | None = ...,
    user_agent_str: str | None = ...,
) -> FileDataContext: ...


@overload
def get_context(
    project_config: DataContextConfig | Mapping | None = ...,
    context_root_dir: None = ...,
    project_root_dir: PathStr = ...,  # If project_root_dir is provided, context_root_dir shouldn't be  # noqa: E501 # FIXME CoP
    runtime_environment: dict | None = ...,
    cloud_base_url: None = ...,
    cloud_access_token: None = ...,
    cloud_organization_id: None = ...,
    cloud_mode: Literal[False] | None = ...,
    user_agent_str: str | None = ...,
    mode: Literal["file"] | None = ...,
) -> FileDataContext: ...


@overload
def get_context(
    project_config: DataContextConfig | Mapping | None = ...,
    context_root_dir: PathStr = ...,  # If context_root_dir is provided, project_root_dir shouldn't be  # noqa: E501 # FIXME CoP
    project_root_dir: None = ...,
    runtime_environment: dict | None = ...,
    cloud_base_url: None = ...,
    cloud_access_token: None = ...,
    cloud_organization_id: None = ...,
    cloud_mode: Literal[False] | None = ...,
    user_agent_str: str | None = ...,
    mode: Literal["file"] | None = ...,
) -> FileDataContext: ...


@overload
def get_context(
    project_config: DataContextConfig | Mapping | None = ...,
    context_root_dir: None = ...,
    project_root_dir: None = ...,
    runtime_environment: dict | None = ...,
    cloud_base_url: str | None = ...,
    cloud_access_token: str | None = ...,
    cloud_organization_id: str | None = ...,
    cloud_mode: Literal[True] = ...,
    user_agent_str: str | None = ...,
    mode: Literal["cloud"] | None = ...,
) -> CloudDataContext: ...


@overload
def get_context(
    project_config: DataContextConfig | Mapping | None = ...,
    context_root_dir: PathStr | None = ...,
    project_root_dir: PathStr | None = ...,
    runtime_environment: dict | None = ...,
    cloud_base_url: str | None = ...,
    cloud_access_token: str | None = ...,
    cloud_organization_id: str | None = ...,
    cloud_mode: bool | None = ...,
    user_agent_str: str | None = ...,
    mode: None = ...,
) -> EphemeralDataContext | FileDataContext | CloudDataContext: ...


@public_api
def get_context(  # noqa: PLR0913 # FIXME CoP
    project_config: DataContextConfig | Mapping | None = None,
    context_root_dir: PathStr | None = None,
    project_root_dir: PathStr | None = None,
    runtime_environment: dict | None = None,
    cloud_base_url: str | None = None,
    cloud_access_token: str | None = None,
    cloud_organization_id: str | None = None,
    cloud_mode: bool | None = None,
    user_agent_str: str | None = None,
    mode: ContextModes | None = None,
) -> AbstractDataContext:
    """Method to return the appropriate Data Context depending on parameters and environment.

    Usage:
        `import great_expectations as gx`

        `my_context = gx.get_context(<insert_your_parameters>)`

    This method returns the appropriate Data Context based on which parameters you've passed and / or your environment configuration:

    - FileDataContext: Configuration stored in a file.
    - EphemeralDataContext: Configuration passed in at runtime.
    - CloudDataContext: Configuration stored in Great Expectations Cloud.

    Read on for more details about each of the Data Context types:

    **FileDataContext:** A Data Context configured via a yaml file. Returned by default if you have no cloud configuration set up and pass no parameters. If you pass context_root_dir, we will look for a great_expectations.yml configuration there. If not we will look at the following locations:

    - Path defined in a GX_HOME environment variable.
    - The current directory.
    - Parent directories of the current directory.

    Relevant parameters

    - project_root_dir: Provide the project root where a GX project exists or will be scaffolded (contains the context root).
    - context_root_dir: Provide an alternative directory to look for GX config.
    - project_config: Optionally override the configuration on disk - only if `context_root_dir` is also provided.
    - runtime_environment: Optionally override specific configuration values.

    **EphemeralDataContext:** A temporary, in-memory Data Context typically used in a pipeline. The default if you pass in only a project_config and have no cloud configuration set up.

    Relevant parameters

    - project_config: Used to configure the Data Context.
    - runtime_environment: Optionally override specific configuration values.

    **CloudDataContext:** A Data Context whose configuration comes from Great Expectations Cloud. The default if you have a cloud configuration set up. Pass `cloud_mode=False` if you have a cloud configuration set up and you do not wish to create a CloudDataContext.

    Cloud configuration can be set up by passing `cloud_*` parameters to `gx.get_context()`, configuring cloud environment variables, or in a great_expectations.conf file.

    Relevant parameters

    - cloud_base_url: Override env var or great_expectations.conf file.
    - cloud_access_token: Override env var or great_expectations.conf file.
    - cloud_organization_id: Override env var or great_expectations.conf file.
    - cloud_mode: Set to True or False to explicitly enable/disable cloud mode.
    - project_config: Optionally override the cloud configuration.
    - runtime_environment: Optionally override specific configuration values.

    Args:
        project_config: In-memory configuration for Data Context.
        context_root_dir (str or pathlib.Path): Path to directory that contains great_expectations.yml file
        project_root_dir (str or pathlib.Path): Path to project root (contains context root with GX config)
        runtime_environment: A dictionary of values can be passed to a DataContext when it is instantiated.
            These values will override both values from the config variables file and
            from environment variables.
        cloud_base_url: url for GX Cloud endpoint.
        cloud_access_token: access_token for GX Cloud account.
        cloud_organization_id: org_id for GX Cloud account.
        cloud_mode: whether to run GX in Cloud mode (default is None).
            If None, cloud mode is assumed if cloud credentials are set up. Set to False to override.
        user_agent_str: Optional string, should be of format <PRODUCT> / <VERSION> <COMMENT>
        mode: which mode to use. One of: ephemeral, file, cloud.
            Note: if mode is specified, cloud_mode is ignored.

    Returns:
        A Data Context. Either a FileDataContext, EphemeralDataContext, or
        CloudDataContext depending on environment and/or
        parameters.

    Raises:
        GXCloudConfigurationError: Cloud mode enabled, but missing configuration.
    """  # noqa: E501 # FIXME CoP
    return project_manager.get_project(
        project_config=project_config,
        context_root_dir=context_root_dir,
        project_root_dir=project_root_dir,
        runtime_environment=runtime_environment,
        cloud_base_url=cloud_base_url,
        cloud_access_token=cloud_access_token,
        cloud_organization_id=cloud_organization_id,
        cloud_mode=cloud_mode,
        user_agent_str=user_agent_str,
        mode=mode,
    )


set_context = project_manager.set_project



import pandas as pd
import pytest

from great_expectations.core.batch import (
    Batch,
    BatchMarkers,
    BatchRequest,
    BatchSpec,
    IDDict,
    LegacyBatchDefinition,
)
from great_expectations.core.batch_spec import RuntimeDataBatchSpec
from great_expectations.core.id_dict import deep_convert_properties_iterable_to_id_dict
from great_expectations.exceptions import InvalidBatchSpecError
from great_expectations.util import convert_to_json_serializable


@pytest.mark.unit
def test_id_dict_structure():
    data: dict = {
        "a0": 1,
        "a1": {
            "b0": "2",
            "b1": [
                "c0",
                (
                    "d0",
                    {
                        "e0": 3,
                        "e1": 4,
                    },
                    {"f0", "f1", "f2"},
                ),
            ],
            "b2": 5,
        },
    }
    nested_id_dictionary: IDDict = deep_convert_properties_iterable_to_id_dict(source=data)
    assert isinstance(nested_id_dictionary, IDDict)
    assert isinstance(nested_id_dictionary["a0"], int)
    assert isinstance(nested_id_dictionary["a1"], IDDict)
    assert isinstance(nested_id_dictionary["a1"]["b0"], str)
    assert isinstance(nested_id_dictionary["a1"]["b1"], list)
    assert isinstance(nested_id_dictionary["a1"]["b1"][0], str)
    assert isinstance(nested_id_dictionary["a1"]["b1"][1], tuple)
    assert isinstance(list(nested_id_dictionary["a1"]["b1"][1])[0], str)
    assert isinstance(list(nested_id_dictionary["a1"]["b1"][1])[1], IDDict)
    assert isinstance(list(nested_id_dictionary["a1"]["b1"][1])[2], set)
    assert isinstance(nested_id_dictionary["a1"]["b2"], int)


@pytest.mark.unit
def test_iddict_is_hashable():
    data_0: dict = {
        "a0": 1,
        "a1": {
            "b0": "2",
            "b1": [
                "c0",
                (
                    "d0",
                    {
                        "e0": 3,
                        "e1": 4,
                    },
                    convert_to_json_serializable(data={"f0", "f1", "f2"}),
                ),
            ],
            "b2": 5,
        },
    }
    data_1: dict = {
        "a0": 0,
        "a1": 1,
    }
    data_2: dict = {
        "b0": 2,
        "b1": 3,
    }
    data_3: dict = {
        "c0": "4",
        "c1": "5",
    }
    # noinspection PyBroadException,PyUnusedLocal
    try:
        # noinspection PyUnusedLocal
        dictionaries_as_set: set = {  # noqa: F841 # FIXME CoP
            deep_convert_properties_iterable_to_id_dict(source=data_0),
            deep_convert_properties_iterable_to_id_dict(source=data_1),
            deep_convert_properties_iterable_to_id_dict(source=data_2),
            deep_convert_properties_iterable_to_id_dict(source=data_3),
        }
    except Exception:
        assert False, "IDDict.__hash__() failed."


@pytest.mark.unit
def test_batch_definition_id():
    # noinspection PyUnusedLocal,PyPep8Naming
    A = LegacyBatchDefinition("A", "a", "aaa", batch_identifiers=IDDict({"id": "A"}))
    print(A.id)

    # noinspection PyUnusedLocal,PyPep8Naming
    B = LegacyBatchDefinition("B", "b", "bbb", batch_identifiers=IDDict({"id": "B"}))
    print(B.id)

    assert A.id != B.id


@pytest.mark.unit
def test_batch_definition_instantiation():
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker,PyUnusedLocal,PyPep8Naming
        A = LegacyBatchDefinition("A", "a", "aaa", {"id": "A"})

    # noinspection PyPep8Naming
    A = LegacyBatchDefinition("A", "a", "aaa", batch_identifiers=IDDict({"id": "A"}))

    print(A.id)


@pytest.mark.unit
def test_batch_definition_equality():
    # noinspection PyUnusedLocal,PyPep8Naming
    A = LegacyBatchDefinition("A", "a", "aaa", batch_identifiers=IDDict({"id": "A"}))

    # noinspection PyUnusedLocal,PyPep8Naming
    B = LegacyBatchDefinition("B", "b", "bbb", batch_identifiers=IDDict({"id": "B"}))

    assert A != B

    # noinspection PyUnusedLocal,PyPep8Naming
    A2 = LegacyBatchDefinition("A", "a", "aaa", batch_identifiers=IDDict({"id": "A"}))

    assert A == A2


@pytest.mark.unit
def test_batch__str__method():
    batch = Batch(
        data=None,
        batch_request=BatchRequest(
            datasource_name="my_datasource",
            data_connector_name="my_data_connector",
            data_asset_name="my_data_asset_name",
        ),
        batch_definition=LegacyBatchDefinition(
            datasource_name="my_datasource",
            data_connector_name="my_data_connector",
            data_asset_name="my_data_asset_name",
            batch_identifiers=IDDict({}),
        ),
        batch_spec=BatchSpec(path="/some/path/some.file"),
        batch_markers=BatchMarkers(ge_load_time="FAKE_LOAD_TIME"),
    )
    print(batch.__str__())

    assert (
        batch.__str__()
        == """{
  "data": "None",
  "batch_request": {
    "datasource_name": "my_datasource",
    "data_connector_name": "my_data_connector",
    "data_asset_name": "my_data_asset_name"
  },
  "batch_definition": {
    "datasource_name": "my_datasource",
    "data_connector_name": "my_data_connector",
    "data_asset_name": "my_data_asset_name",
    "batch_identifiers": {}
  },
  "batch_spec": {
    "path": "/some/path/some.file"
  },
  "batch_markers": {
    "ge_load_time": "FAKE_LOAD_TIME"
  }
}"""
    )


@pytest.mark.unit
def test_batch_request_instantiation():
    BatchRequest(
        datasource_name="A",
        data_connector_name="a",
        data_asset_name="aaa",
        data_connector_query={"id": "A"},
    )

    BatchRequest("A", "a", "aaa", {"id": "A"})

    # with pytest.raises(TypeError):
    #     BatchRequest(
    #         "A",
    #         "a",
    #         "aaa",
    #         IDDict({
    #             "id": "A"
    #         })
    #     )

    # No data_source_name specified
    with pytest.raises(TypeError):
        # noinspection PyArgumentList
        BatchRequest(
            data_connector_name="a",
            data_asset_name="aaa",
            data_connector_query={"id": "A"},
        )

    # No data_source_name and data_connector_name specified
    with pytest.raises(TypeError):
        # noinspection PyArgumentList
        BatchRequest(data_asset_name="aaa", data_connector_query={"id": "A"})

    # No data_source_name and data_connector_name and data_asset_name specified
    with pytest.raises(TypeError):
        # noinspection PyArgumentList
        BatchRequest(data_connector_query={"id": "A"})

    BatchRequest(datasource_name="A", data_connector_name="a", data_asset_name="aaa")


# noinspection PyPep8Naming
@pytest.mark.unit
def test_RuntimeDataBatchSpec():
    with pytest.raises(InvalidBatchSpecError):
        RuntimeDataBatchSpec()

    RuntimeDataBatchSpec({"batch_data": pd.DataFrame({"x": range(10)})})

    RuntimeDataBatchSpec(
        batch_data="we don't check types yet",
    )

    RuntimeDataBatchSpec(
        {
            "batch_data": "we don't check types yet",
        }
    )



from __future__ import annotations

import os
import uuid

from ruamel.yaml import YAML
from ruamel.yaml.compat import StringIO

from great_expectations.data_context.types.base import DataContextConfigDefaults


class YAMLToString(YAML):
    """
    Get yaml dump as a string: https://yaml.readthedocs.io/en/latest/example.html#output-of-dump-as-a-string
    """

    def dump(self, data, stream=None, **kw):  # type: ignore[explicit-override] # FIXME
        inefficient = False
        if not stream:
            inefficient = True
            stream = StringIO()
        YAML.dump(self, data, stream, **kw)
        if inefficient:
            return stream.getvalue()


yaml = YAMLToString()
yaml.indent(mapping=2, sequence=4, offset=4)
yaml.default_flow_style = False

# TODO: maybe bring params in via f-strings from base.ConfigDefaults or whatever
#  I end up using for the base level configs. Specifically PROJECT_OPTIONAL_CONFIG_COMMENT
#  and PROJECT_HELP_COMMENT

PROJECT_HELP_COMMENT = f"""
# Welcome to Great Expectations! Always know what to expect from your data.
#
# Here you can define datasources, batch kwargs generators, integrations and
# more. This file is intended to be committed to your repo. For help with
# configuration please:
#   - Read our docs: https://docs.greatexpectations.io/docs/guides/connecting_to_your_data/connect_to_data_overview/#2-configure-your-datasource
#   - Join our slack channel: http://greatexpectations.io/slack

# config_version refers to the syntactic version of this config file, and is used in maintaining backwards compatibility
# It is auto-generated and usually does not need to be changed.
config_version: {DataContextConfigDefaults.DEFAULT_CONFIG_VERSION.value}
"""  # noqa: E501 # FIXME CoP

CONFIG_VARIABLES_INTRO = """
# This config file supports variable substitution which enables: 1) keeping
# secrets out of source control & 2) environment-based configuration changes
# such as staging vs prod.
#
# When GX encounters substitution syntax (like `my_key: ${my_value}` or
# `my_key: $my_value`) in the great_expectations.yml file, it will attempt
# to replace the value of `my_key` with the value from an environment
# variable `my_value` or a corresponding key read from this config file,
# which is defined through the `config_variables_file_path`.
# Environment variables take precedence over variables defined here.
#
# Substitution values defined here can be a simple (non-nested) value,
# nested value such as a dictionary, or an environment variable (i.e. ${ENV_VAR})
#
#
# https://docs.greatexpectations.io/docs/guides/setup/configuring_data_contexts/how_to_configure_credentials

"""

CONFIG_VARIABLES_TEMPLATE = f"{CONFIG_VARIABLES_INTRO}instance_id: {uuid.uuid4()!s}{os.linesep}"

# Create yaml strings
# NOTE: .replace("\n", "\n  ")[:-2] is a hack to indent all lines two spaces,
# and remove the inserted final two spaces.
EXPECTATIONS_STORE_STRING = yaml.dump(
    {"expectations_store": DataContextConfigDefaults.DEFAULT_STORES.value["expectations_store"]}
).replace("\n", "\n  ")[:-2]
VALIDATIONS_STORE_STRING = yaml.dump(
    {
        "validation_results_store": DataContextConfigDefaults.DEFAULT_STORES.value[
            "validation_results_store"
        ]
    }
).replace("\n", "\n  ")[:-2]
CHECKPOINT_STORE_STRING = yaml.dump(
    {"checkpoint_store": DataContextConfigDefaults.DEFAULT_STORES.value["checkpoint_store"]}
).replace("\n", "\n  ")[:-2]
VALIDATION_DEFINITION_STORE_STRING = yaml.dump(
    {
        "validation_definition_store": DataContextConfigDefaults.DEFAULT_STORES.value[
            "validation_definition_store"
        ]
    }
).replace("\n", "\n  ")[:-2]

PROJECT_OPTIONAL_CONFIG_COMMENT = (
    CONFIG_VARIABLES_INTRO
    + f"""
config_variables_file_path: {DataContextConfigDefaults.DEFAULT_CONFIG_VARIABLES_FILEPATH.value}

# The plugins_directory will be added to your python path for custom modules
# used to override and extend Great Expectations.
plugins_directory: {DataContextConfigDefaults.DEFAULT_PLUGINS_DIRECTORY.value}

stores:
# Stores are configurable places to store things like Expectations, Validations
# Data Docs, and more. These are for advanced users only - most users can simply
# leave this section alone.
  {EXPECTATIONS_STORE_STRING}
  {VALIDATIONS_STORE_STRING}
  {CHECKPOINT_STORE_STRING}
  {VALIDATION_DEFINITION_STORE_STRING}
expectations_store_name: expectations_store
validation_results_store_name: validation_results_store
checkpoint_store_name: checkpoint_store

data_docs_sites:
  # Data Docs make it simple to visualize data quality in your project. These
  # include Expectations, Validations & Profiles. The are built for all
  # Datasources from JSON artifacts in the local repo including validations &
  # profiles from the uncommitted directory. Read more at https://docs.greatexpectations.io/docs/terms/data_docs
  local_site:
    class_name: SiteBuilder
    # set to false to hide how-to buttons in Data Docs
    show_how_to_buttons: true
    store_backend:
        class_name: TupleFilesystemStoreBackend
        base_directory: uncommitted/data_docs/local_site/
    site_index_builder:
        class_name: DefaultSiteIndexBuilder
"""
)


PROJECT_TEMPLATE_USAGE_STATISTICS_ENABLED = PROJECT_HELP_COMMENT + PROJECT_OPTIONAL_CONFIG_COMMENT



from __future__ import annotations

import datetime as dt
import logging
from typing import TYPE_CHECKING

from great_expectations.compatibility.typing_extensions import override
from great_expectations.render.renderer.renderer import Renderer

if TYPE_CHECKING:
    from great_expectations.checkpoint.checkpoint import CheckpointResult
    from great_expectations.core.expectation_validation_result import (
        ExpectationSuiteValidationResult,
    )
    from great_expectations.data_context.types.resource_identifiers import (
        ValidationResultIdentifier,
    )

logger = logging.getLogger(__name__)


class MicrosoftTeamsRenderer(Renderer):
    """
    Responsible for formatting validation results and data docs links into a Microsoft Teams webhook
    message payload.

    Relevant links/documentation:
        * Payload schema: https://adaptivecards.io/explorer/
        * Interactive UI editor: https://adaptivecards.io/designer/

    """

    _MICROSOFT_TEAMS_CONTENT_TYPE = "application/vnd.microsoft.card.adaptive"
    _MICROSOFT_TEAMS_SCHEMA_URL = "http://adaptivecards.io/schemas/adaptive-card.json"
    _MICROSOFT_TEAMS_SCHEMA_VERSION = "1.5"
    _GX_LOGO_URL = "https://www.greatexpectations.io/image/gx-logo-mark-400"
    _SUCCESS_EMOJI = "✅"
    _FAILURE_EMOJI = "❌"
    _NO_VALUE_PLACEHOLDER = "--"

    @override
    def render(
        self,
        checkpoint_result: CheckpointResult,
        data_docs_pages: dict[ValidationResultIdentifier, dict[str, str]] | None = None,
    ) -> dict:
        blocks: list[dict] = []

        blocks.append(
            self._build_header_block(
                checkpoint_name=checkpoint_result.name,
                success=checkpoint_result.success or False,
                run_time=checkpoint_result.run_id.run_time,
            )
        )

        for idx, (validation_result_identifier, validation_result) in enumerate(
            checkpoint_result.run_results.items(), start=1
        ):
            validation_blocks = self._build_validation_result_blocks(
                idx=idx,
                total_validation_count=len(checkpoint_result.run_results),
                validation_result_identifier=validation_result_identifier,
                validation_result=validation_result,
            )
            blocks.extend(validation_blocks)

        return self._build_payload(blocks=blocks, data_docs_pages=data_docs_pages)

    def _build_header_block(
        self, checkpoint_name: str, success: bool, run_time: dt.datetime
    ) -> dict:
        success_text = (
            f"Success {self._SUCCESS_EMOJI}" if success else f"Failure {self._FAILURE_EMOJI}"
        )
        return {
            "type": "ColumnSet",
            "columns": [
                {
                    "type": "Column",
                    "items": [
                        {
                            "type": "Image",
                            "url": self._GX_LOGO_URL,
                            "altText": checkpoint_name,
                            "size": "small",
                        }
                    ],
                    "width": "auto",
                },
                {
                    "type": "Column",
                    "items": [
                        {
                            "type": "TextBlock",
                            "weight": "bolder",
                            "text": f"{checkpoint_name} - {success_text}",
                            "wrap": True,
                        },
                        {
                            "type": "TextBlock",
                            "spacing": "None",
                            "text": f"Ran {run_time}",
                            "isSubtle": True,
                            "wrap": True,
                        },
                    ],
                    "width": "stretch",
                },
            ],
        }

    def _build_validation_result_blocks(
        self,
        idx: int,
        total_validation_count: int,
        validation_result_identifier: ValidationResultIdentifier,
        validation_result: ExpectationSuiteValidationResult,
    ) -> list[dict]:
        success = validation_result.success
        success_text = self._SUCCESS_EMOJI if success else self._FAILURE_EMOJI
        color = "Good" if success else "Attention"

        blocks = [
            {
                "type": "TextBlock",
                "text": f"Validation Result ({idx} of {total_validation_count}) {success_text}",
                "wrap": True,
                "weight": "Bolder",
                "style": "columnHeader",
                "isSubtle": True,
                "color": color,
            },
            {
                "type": "FactSet",
                "facts": self._build_validation_result_facts(
                    validation_result_identifier=validation_result_identifier,
                    validation_result=validation_result,
                ),
                "separator": True,
            },
        ]

        if validation_result.result_url:
            blocks.append(
                {
                    "type": "ActionSet",
                    "actions": [
                        {
                            "type": "Action.OpenUrl",
                            "title": "View Result",
                            "url": validation_result.result_url,
                        }
                    ],
                }
            )

        return blocks

    def _build_validation_result_facts(
        self,
        validation_result_identifier: ValidationResultIdentifier,
        validation_result: ExpectationSuiteValidationResult,
    ) -> list[dict]:
        asset_name = validation_result.asset_name or self._NO_VALUE_PLACEHOLDER
        suite_name = validation_result.suite_name
        run_name = validation_result_identifier.run_id.run_name or self._NO_VALUE_PLACEHOLDER
        n_checks_succeeded = validation_result.statistics["successful_expectations"]
        n_checks = validation_result.statistics["evaluated_expectations"]
        check_details_text = f"*{n_checks_succeeded}* of *{n_checks}* Expectations were met"

        return [
            {"title": "Data Asset name: ", "value": asset_name},
            {"title": "Suite name: ", "value": suite_name},
            {"title": "Run name: ", "value": run_name},
            {"title": "Summary:", "value": check_details_text},
        ]

    def _get_data_docs_page_links(
        self, data_docs_pages: dict[ValidationResultIdentifier, dict[str, str]] | None
    ) -> list[str]:
        links: list[str] = []

        if not data_docs_pages:
            return links

        for data_docs_page in data_docs_pages.values():
            for docs_link_key, docs_link in data_docs_page.items():
                if docs_link_key == "class":
                    continue

                links.append(docs_link)

        return links

    def _build_payload(
        self,
        blocks: list[dict],
        data_docs_pages: dict[ValidationResultIdentifier, dict[str, str]] | None,
    ) -> dict:
        data_docs_page_links = self._get_data_docs_page_links(data_docs_pages)
        actions = [
            # We would normally use Action.OpenUrl here, but Teams does not support
            # non HTTP/HTTPS URI schemes. As Data Docs utilze file:///, we use Action.ShowCard
            # to display the link in a card.
            {
                "type": "Action.ShowCard",
                "title": "View Data Docs URL",
                "card": {
                    "type": "AdaptiveCard",
                    "body": [
                        {
                            "type": "TextBlock",
                            "text": link,
                            "wrap": True,
                        },
                    ],
                    "$schema": self._MICROSOFT_TEAMS_SCHEMA_URL,
                },
            }
            for link in data_docs_page_links
        ]

        return {
            "type": "message",
            "attachments": [
                {
                    "contentType": self._MICROSOFT_TEAMS_CONTENT_TYPE,
                    "content": {
                        "type": "AdaptiveCard",
                        "$schema": self._MICROSOFT_TEAMS_SCHEMA_URL,
                        "version": self._MICROSOFT_TEAMS_SCHEMA_VERSION,
                        "body": blocks,
                        "actions": actions,
                    },
                }
            ],
        }



from __future__ import annotations

"""Contains general abstract or base classes used across configuration objects."""
from abc import ABC
from pprint import pformat as pf
from typing import Optional

from marshmallow.decorators import post_dump
from marshmallow.schema import Schema

from great_expectations.compatibility.typing_extensions import override
from great_expectations.types import SerializableDictDot


class AbstractConfig(ABC, SerializableDictDot):
    """Abstract base class for Config objects. Sets the fields that must be included on a Config."""

    def __init__(self, id: Optional[str] = None, name: Optional[str] = None) -> None:
        self.id = id
        self.name = name
        super().__init__()

    @override
    def __repr__(self) -> str:
        return pf(self.to_dict(), indent=2, sort_dicts=True)

    @classmethod
    def _dict_round_trip(cls, schema: Schema, target: dict) -> dict:
        """
        Round trip a dictionary with a schema so that validation and serialization logic is applied.

        Example: Loading a config with a `id_` field but serializing it as `id`.
        """
        _loaded = schema.load(target)
        _config = cls(**_loaded)
        return _config.to_json_dict()


class AbstractConfigSchema(Schema):
    REMOVE_KEYS_IF_NONE = ["id", "name"]

    @post_dump
    def filter_none(self, data: dict, **kwargs) -> dict:
        return {
            key: value
            for key, value in data.items()
            if key not in AbstractConfigSchema.REMOVE_KEYS_IF_NONE or value is not None
        }



from __future__ import annotations

import copy
import logging
import re
import sre_constants
import sre_parse
from abc import abstractmethod
from collections import defaultdict
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Set, Tuple, Union

from great_expectations.compatibility.typing_extensions import override
from great_expectations.core import IDDict
from great_expectations.core.batch import LegacyBatchDefinition
from great_expectations.core.batch_spec import BatchSpec, PathBatchSpec
from great_expectations.datasource.fluent.batch_identifier_util import make_batch_identifier
from great_expectations.datasource.fluent.constants import _DATA_CONNECTOR_NAME, MATCH_ALL_PATTERN
from great_expectations.datasource.fluent.data_connector import (
    DataConnector,
)
from great_expectations.datasource.fluent.data_connector.batch_filter import (
    BatchFilter,
    build_batch_filter,
)
from great_expectations.datasource.fluent.data_connector.regex_parser import (
    RegExParser,
)

if TYPE_CHECKING:
    from typing import DefaultDict

    from great_expectations.alias_types import PathStr
    from great_expectations.core.partitioners import FileNamePartitioner
    from great_expectations.datasource.fluent import BatchRequest

logger = logging.getLogger(__name__)


class MissingFilePathTemplateMapFnError(ValueError):
    def __init__(self):
        super().__init__(
            "Converting file paths to fully-qualified object references for "
            f"`{self.__class__.__name__}` requires "
            "`file_path_template_map_fn: Callable` to be set."
        )


class FilePathDataConnector(DataConnector):
    """The base class for Data Connectors designed to access filesystem-like data.

    This can include traditional, disk-based filesystems or object stores such as S3, GCS, or ABS.

    See the `DataConnector` base class for more information on the role of Data Connectors.

    Note that `FilePathDataConnector` is not meant to be used on its own, but extended.

    Args:
        datasource_name: The name of the Datasource associated with this DataConnector instance
        data_asset_name: The name of the DataAsset using this DataConnector instance
    """

    FILE_PATH_BATCH_SPEC_KEY = "path"

    def __init__(
        self,
        datasource_name: str,
        data_asset_name: str,
        unnamed_regex_group_prefix: str = "batch_request_param_",
        file_path_template_map_fn: Optional[Callable] = None,
        whole_directory_path_override: PathStr | None = None,
    ) -> None:
        super().__init__(
            datasource_name=datasource_name,
            data_asset_name=data_asset_name,
        )

        self._unnamed_regex_group_prefix: str = unnamed_regex_group_prefix

        self._file_path_template_map_fn: Optional[Callable] = file_path_template_map_fn

        # allow callers to always treat entire directory as single asset
        self._whole_directory_path_override = whole_directory_path_override

        # This is a dictionary which maps data_references onto batch_requests.
        self._data_references_cache: DefaultDict[
            re.Pattern, Dict[str, List[LegacyBatchDefinition] | None]
        ] = defaultdict(dict)

    # Interface Method
    @override
    def get_batch_definition_list(self, batch_request: BatchRequest) -> List[LegacyBatchDefinition]:
        """
        Retrieve batch_definitions and that match batch_request.

        First retrieves all batch_definitions that match batch_request
            - if batch_request also has a batch_filter, then select batch_definitions that match batch_filter.

        Args:
            batch_request (BatchRequest): BatchRequest (containing previously validated attributes) to process

        Returns:
            A list of BatchDefinition objects that match BatchRequest

        """  # noqa: E501 # FIXME CoP
        legacy_batch_definition_list: List[LegacyBatchDefinition] = (
            self._get_unfiltered_batch_definition_list(batch_request=batch_request)
        )

        data_connector_query_dict: dict[str, dict | slice] = {}
        if batch_request.options:
            data_connector_query_dict.update(
                {
                    "batch_filter_parameters": {
                        key: value
                        for key, value in batch_request.options.items()
                        if value is not None
                    }
                }
            )

        data_connector_query_dict.update({"index": batch_request.batch_slice})

        batch_filter_obj: BatchFilter = build_batch_filter(
            data_connector_query_dict=data_connector_query_dict  # type: ignore[arg-type] # FIXME CoP
        )
        legacy_batch_definition_list = batch_filter_obj.select_from_data_connector_query(
            batch_definition_list=legacy_batch_definition_list
        )

        return legacy_batch_definition_list

    @override
    def build_batch_spec(self, batch_definition: LegacyBatchDefinition) -> PathBatchSpec:
        """
        Build BatchSpec from batch_definition by calling DataConnector's build_batch_spec function.

        Args:
            batch_definition (LegacyBatchDefinition): to be used to build batch_spec

        Returns:
            BatchSpec built from batch_definition
        """
        batch_spec: BatchSpec = super().build_batch_spec(batch_definition=batch_definition)
        return PathBatchSpec(batch_spec)

    # Interface Method
    @override
    def get_data_reference_count(self) -> int:
        # todo: in the world of BatchDefinition, this method must accept a BatchRequest.
        #       In the meantime, we fall back to a regex that matches everything.
        regex = self._preprocess_batching_regex(MATCH_ALL_PATTERN)
        data_references = self._get_data_references_cache(batching_regex=regex)
        return len(data_references)

    # Interface Method
    @override
    def get_matched_data_references(self, regex: re.Pattern | None = None) -> List[str]:
        """
        Returns the list of data_references matched by configuration by looping through items in
        _data_references_cache and returning data_references that have an associated data_asset.

        Returns:
            list of data_references that are matched by configuration.
        """
        if regex:
            regex = self._preprocess_batching_regex(regex)
        return self._get_data_references(matched=True, regex=regex)

    # Interface Method
    @override
    def get_matched_data_reference_count(self) -> int:
        """
        Returns the list of matched data_references known by this DataConnector from its _data_references_cache

        Returns:
            number of matched data_references known by this DataConnector.
        """  # noqa: E501 # FIXME CoP
        return len(self.get_matched_data_references())

    # Interface Method
    @override
    def get_unmatched_data_references(self) -> List[str]:
        """
        Returns the list of data_references unmatched by configuration by looping through items in
        _data_references_cache and returning data_references that do not have an associated data_asset.

        Returns:
            list of data_references that are not matched by configuration.
        """  # noqa: E501 # FIXME CoP
        return self._get_data_references(matched=False)

    # Interface Method
    @override
    def get_unmatched_data_reference_count(self) -> int:
        """
        Returns the list of unmatched data_references known by this DataConnector from its _data_references_cache

        Returns:
            number of unmached data_references known by this DataConnector.
        """  # noqa: E501 # FIXME CoP
        return len(self.get_unmatched_data_references())

    def _get_unfiltered_batch_definition_list(
        self, batch_request: BatchRequest[FileNamePartitioner]
    ) -> list[LegacyBatchDefinition]:
        """Get all batch definitions for all files from a data connector
         using the supplied batch request.

        Args:
            batch_request: Specifies which batch definitions to get from data connector.

        Returns:
            A list of batch definitions from the data connector based on the batch request.
        """
        # this class is overloaded with two separate implementations:
        if self._whole_directory_path_override:
            return self._get_directory_batch_definition_list(batch_request=batch_request)
        else:
            return self._get_file_batch_definition_list(batch_request=batch_request)

    def _get_file_batch_definition_list(
        self, batch_request: BatchRequest
    ) -> list[LegacyBatchDefinition]:
        # Use a combination of a list and set to preserve iteration order
        batch_definition_list: list[LegacyBatchDefinition] = list()
        batch_definition_set = set()
        if batch_request.partitioner:
            batching_regex = self._preprocess_batching_regex(batch_request.partitioner.regex)
        else:
            # all batch requests coming from the V1 API should have a regex; to support legacy code
            # we fall back to the MATCH_ALL_PATTERN if it's missing.
            batching_regex = self._preprocess_batching_regex(MATCH_ALL_PATTERN)
        for batch_definition in self._get_batch_definitions(batching_regex=batching_regex):
            if (
                self._batch_definition_matches_batch_request(
                    batch_definition=batch_definition, batch_request=batch_request
                )
                and batch_definition not in batch_definition_set
            ):
                batch_definition_list.append(batch_definition)
                batch_definition_set.add(batch_definition)

        return batch_definition_list

    def _get_directory_batch_definition_list(
        self, batch_request: BatchRequest
    ) -> list[LegacyBatchDefinition]:
        data_directory = self._whole_directory_path_override
        batch_definition = LegacyBatchDefinition(
            datasource_name=self._datasource_name,
            data_connector_name=_DATA_CONNECTOR_NAME,
            data_asset_name=self._data_asset_name,
            batch_identifiers=make_batch_identifier({"path": data_directory}),
        )
        return [batch_definition]

    def _get_data_references(self, matched: bool, regex: re.Pattern | None = None) -> List[str]:
        """
        Returns the list of data_references unmatched by configuration by looping through items in
        _data_references_cache and returning data_references that do not have an associated data_asset.

        Returns:
            list of data_references that are not matched by configuration.
        """  # noqa: E501 # FIXME CoP
        if not regex:
            regex = self._preprocess_batching_regex(MATCH_ALL_PATTERN)

        def _matching_criterion(
            batch_definition_list: Union[List[LegacyBatchDefinition], None],
        ) -> bool:
            return (
                (batch_definition_list is not None) if matched else (batch_definition_list is None)
            )

        data_reference_mapped_element: Tuple[str, Union[List[LegacyBatchDefinition], None]]
        data_references = self._get_data_references_cache(batching_regex=regex)
        unmatched_data_references: List[str] = list(
            dict(
                filter(
                    lambda data_reference_mapped_element: _matching_criterion(
                        batch_definition_list=data_reference_mapped_element[1]
                    ),
                    data_references.items(),
                )
            ).keys()
        )
        return unmatched_data_references

    # Interface Method
    @override
    def _generate_batch_spec_parameters_from_batch_definition(
        self, batch_definition: LegacyBatchDefinition
    ) -> dict:
        """
        This interface method examines "BatchDefinition" object and converts it to exactly one "data_reference" handle,
        based on partitioning behavior of given subclass (e.g., Regular Expressions for file path based DataConnector
        implementations).  Type of "data_reference" is storage dependent.  This method is then used to create storage
        system specific "BatchSpec" parameters for retrieving "Batch" of data.

        Args:
            batch_definition: input "BatchDefinition" object

        Returns:
            dict -- dictionary of "BatchSpec" properties
        """  # noqa: E501 # FIXME CoP
        # this class is overloaded with two separate implementations:
        if self._whole_directory_path_override:
            return self._get_batch_spec_params_directory(batch_definition=batch_definition)
        else:
            return self._get_batch_spec_params_file(batch_definition=batch_definition)

    def _get_batch_spec_params_file(self, batch_definition: LegacyBatchDefinition) -> dict:
        """File specific implementation of batch spec parameters"""
        if not batch_definition.batching_regex:
            raise RuntimeError("BatchDefinition must contain a batching_regex.")  # noqa: TRY003 # FIXME CoP

        batching_regex = batch_definition.batching_regex

        regex_parser = RegExParser(
            regex_pattern=batching_regex,
            unnamed_regex_group_prefix=self._unnamed_regex_group_prefix,
        )
        group_names: List[str] = regex_parser.group_names()
        path: str = map_batch_definition_to_data_reference_string_using_regex(
            batch_definition=batch_definition,
            regex_pattern=batching_regex,
            group_names=group_names,
        )
        if not path:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                f"""No data reference for data asset name "{batch_definition.data_asset_name}" matches the given
        batch identifiers {batch_definition.batch_identifiers} from batch definition {batch_definition}.
        """  # noqa: E501 # FIXME CoP
            )

        path = self._get_full_file_path(path=path)

        return {FilePathDataConnector.FILE_PATH_BATCH_SPEC_KEY: path}

    def _get_batch_spec_params_directory(self, batch_definition: LegacyBatchDefinition) -> dict:
        """Directory specific implementation of batch spec parameters"""
        path = self._get_full_file_path(path=str(self._whole_directory_path_override))

        return {FilePathDataConnector.FILE_PATH_BATCH_SPEC_KEY: path}

    def _preprocess_batching_regex(self, regex: re.Pattern) -> re.Pattern:
        """Add the FILE_PATH_BATCH_SPEC_KEY group to regex if not already present."""
        regex_parser = RegExParser(
            regex_pattern=regex,
            unnamed_regex_group_prefix=self._unnamed_regex_group_prefix,
        )
        group_names: List[str] = regex_parser.group_names()
        if FilePathDataConnector.FILE_PATH_BATCH_SPEC_KEY not in group_names:
            pattern: str = regex.pattern
            pattern = f"(?P<{FilePathDataConnector.FILE_PATH_BATCH_SPEC_KEY}>{pattern})"
            regex = re.compile(pattern)

        return regex

    def _get_data_references_cache(
        self, batching_regex: re.Pattern
    ) -> Dict[str, List[LegacyBatchDefinition] | None]:
        """Access a map where keys are data references and values are LegacyBatchDefinitions."""

        batch_definitions = self._data_references_cache[batching_regex]
        if batch_definitions:
            return batch_definitions

        # Cache was empty so we need to calculate BatchDefinitions
        for data_reference in self.get_data_references():
            batch_definition = self._build_batch_definition(
                data_reference=data_reference, batching_regex=batching_regex
            )
            if batch_definition:
                # storing these as a list seems unnecessary; in this implementation
                # there can only be one or zero BatchDefinitions per data reference
                batch_definitions[data_reference] = [batch_definition]
            else:
                batch_definitions[data_reference] = None

        return batch_definitions

    def _get_batch_definitions(self, batching_regex: re.Pattern) -> List[LegacyBatchDefinition]:
        batch_definition_map = self._get_data_references_cache(batching_regex=batching_regex)
        batch_definitions = [
            batch_definitions[0]
            for batch_definitions in batch_definition_map.values()
            if batch_definitions is not None
        ]
        return batch_definitions

    def _build_batch_definition(
        self, data_reference: str, batching_regex: re.Pattern
    ) -> LegacyBatchDefinition | None:
        batch_identifiers = self._build_batch_identifiers(
            data_reference=data_reference, batching_regex=batching_regex
        )
        if batch_identifiers is None:
            return None

        from great_expectations.core.batch import LegacyBatchDefinition

        return LegacyBatchDefinition(
            datasource_name=self._datasource_name,
            data_connector_name=_DATA_CONNECTOR_NAME,
            data_asset_name=self._data_asset_name,
            batch_identifiers=batch_identifiers,
            batching_regex=batching_regex,
        )

    def _build_batch_identifiers(
        self, data_reference: str, batching_regex: re.Pattern
    ) -> Optional[IDDict]:
        regex_parser = RegExParser(
            regex_pattern=batching_regex,
            unnamed_regex_group_prefix=self._unnamed_regex_group_prefix,
        )
        matches: Optional[re.Match] = regex_parser.get_matches(target=data_reference)
        if matches is None:
            return None

        num_all_matched_group_values: int = regex_parser.get_num_all_matched_group_values()

        # Check for `(?P<name>)` named group syntax
        defined_group_name_to_group_index_mapping: Dict[str, int] = (
            regex_parser.get_named_group_name_to_group_index_mapping()
        )
        defined_group_name_indexes: Set[int] = set(
            defined_group_name_to_group_index_mapping.values()
        )
        defined_group_name_to_group_value_mapping: Dict[str, str] = matches.groupdict()

        all_matched_group_values: List[str] = list(matches.groups())

        assert len(all_matched_group_values) == num_all_matched_group_values

        group_name_to_group_value_mapping: Dict[str, str] = copy.deepcopy(
            defined_group_name_to_group_value_mapping
        )

        idx: int
        group_idx: int
        matched_group_value: str
        for idx, matched_group_value in enumerate(all_matched_group_values):
            group_idx = idx + 1
            if group_idx not in defined_group_name_indexes:
                group_name: str = f"{self._unnamed_regex_group_prefix}{group_idx}"
                group_name_to_group_value_mapping[group_name] = matched_group_value

        batch_identifiers = make_batch_identifier(group_name_to_group_value_mapping)

        return batch_identifiers

    @abstractmethod
    def _get_full_file_path(self, path: str) -> str:
        pass


def map_batch_definition_to_data_reference_string_using_regex(
    batch_definition: LegacyBatchDefinition,
    regex_pattern: re.Pattern,
    group_names: List[str],
) -> str:
    if not isinstance(batch_definition, LegacyBatchDefinition):
        raise TypeError("batch_definition is not of an instance of type BatchDefinition")  # noqa: TRY003 # FIXME CoP

    data_asset_name: str = batch_definition.data_asset_name
    batch_identifiers: IDDict = batch_definition.batch_identifiers
    data_reference: str = convert_batch_identifiers_to_data_reference_string_using_regex(
        batch_identifiers=batch_identifiers,
        regex_pattern=regex_pattern,
        group_names=group_names,
        data_asset_name=data_asset_name,
    )
    return data_reference


def convert_batch_identifiers_to_data_reference_string_using_regex(
    batch_identifiers: IDDict,
    regex_pattern: re.Pattern,
    group_names: List[str],
    data_asset_name: Optional[str] = None,
) -> str:
    if not isinstance(batch_identifiers, IDDict):
        raise TypeError("batch_identifiers is not an instance of type IDDict")  # noqa: TRY003 # FIXME CoP

    template_arguments: dict = copy.deepcopy(batch_identifiers)
    if data_asset_name is not None:
        template_arguments["data_asset_name"] = data_asset_name

    filepath_template: str = _invert_regex_to_data_reference_template(
        regex_pattern=regex_pattern,
        group_names=group_names,
    )
    converted_string: str = filepath_template.format(**template_arguments)

    return converted_string


def _invert_regex_to_data_reference_template(  # noqa: C901 #  too complex
    regex_pattern: re.Pattern | str,
    group_names: List[str],
) -> str:
    r"""Create a string template based on a regex and corresponding list of group names.

    For example:

        filepath_template = _invert_regex_to_data_reference_template(
            regex_pattern=r"^(.+)_(\d+)_(\d+)\.csv$",
            group_names=["name", "timestamp", "price"],
        )
        filepath_template
        >> "{name}_{timestamp}_{price}.csv"

    Such templates are useful because they can be populated using string substitution:

        filepath_template.format(**{
            "name": "user_logs",
            "timestamp": "20200101",
            "price": "250",
        })
        >> "user_logs_20200101_250.csv"


    NOTE Abe 20201017: This method is almost certainly still brittle. I haven't exhaustively mapped the OPCODES in sre_constants
    """  # noqa: E501 # FIXME CoP
    data_reference_template: str = ""
    group_name_index: int = 0

    num_groups = len(group_names)

    if isinstance(regex_pattern, re.Pattern):
        regex_pattern = regex_pattern.pattern

    # print("-"*80)
    parsed_sre = sre_parse.parse(str(regex_pattern))
    for parsed_sre_tuple, char in zip(parsed_sre, list(str(regex_pattern))):  # type: ignore[call-overload] # FIXME CoP
        token, value = parsed_sre_tuple
        if token == sre_constants.LITERAL:
            # Transcribe the character directly into the template
            data_reference_template += chr(value)
        elif token == sre_constants.SUBPATTERN:
            if not (group_name_index < num_groups):
                break
            # Replace the captured group with "{next_group_name}" in the template
            data_reference_template += f"{{{group_names[group_name_index]}}}"
            group_name_index += 1
        elif token in [
            sre_constants.MAX_REPEAT,
            sre_constants.IN,
            sre_constants.BRANCH,
            sre_constants.ANY,
        ]:
            if group_names:
                # Replace the uncaptured group a wildcard in the template
                data_reference_template += "*"
            else:
                # Don't assume that a `.` in a filename should be a star glob
                data_reference_template += char
        elif token in [
            sre_constants.AT,
            sre_constants.ASSERT_NOT,
            sre_constants.ASSERT,
        ]:
            pass
        else:
            raise ValueError(f"Unrecognized regex token {token} in regex pattern {regex_pattern}.")  # noqa: TRY003 # FIXME CoP

    # Collapse adjacent wildcards into a single wildcard
    data_reference_template: str = re.sub("\\*+", "*", data_reference_template)  # type: ignore[no-redef] # FIXME CoP

    return data_reference_template


def sanitize_prefix_for_gcs_and_s3(text: str) -> str:
    """
    Takes in a given user-prefix and cleans it to work with file-system traversal methods
    (i.e. add '/' to the end of a string meant to represent a directory)

    Customized for S3 paths, ignoring the path separator used by the host OS
    """
    text = text.strip()
    if not text:
        return text

    path_parts = text.split("/")
    if not path_parts:  # Empty prefix
        return text

    if "." in path_parts[-1]:  # File, not folder
        return text

    # Folder, should have trailing /
    return f"{text.rstrip('/')}/"



from __future__ import annotations

import copy
import logging
from decimal import Decimal
from typing import TYPE_CHECKING

import pandas as pd
import pytest

from great_expectations.compatibility import pyspark
from great_expectations.core.batch import RuntimeBatchRequest
from great_expectations.data_context.types.base import (
    AbstractConfig,
    AssetConfig,
    DataConnectorConfig,
    assetConfigSchema,
    dataConnectorConfigSchema,
)
from great_expectations.util import (
    convert_to_json_serializable,
    deep_filter_properties_iterable,
    requires_lossy_conversion,
)

if TYPE_CHECKING:
    from _pytest.fixtures import FixtureRequest
    from marshmallow import Schema


@pytest.fixture
def spark_schema(spark_session: pyspark.SparkSession) -> pyspark.types.StructType:
    return pyspark.types.StructType(
        [
            pyspark.types.StructField("a", pyspark.types.IntegerType(), True, None),
            pyspark.types.StructField("b", pyspark.types.IntegerType(), True, None),
        ]
    )


@pytest.fixture
def data_connector_config_spark(spark_session) -> DataConnectorConfig:
    return DataConnectorConfig(
        class_name="ConfiguredAssetFilesystemDataConnector",
        module_name="great_expectations.datasource.data_connector.configured_asset_filesystem_data_connector",
        batch_spec_passthrough={
            "reader_options": {"header": True},
        },
    )


@pytest.fixture
def datas_connector_config_with_schema_spark(spark_session, spark_schema) -> DataConnectorConfig:
    return DataConnectorConfig(
        class_name="ConfiguredAssetFilesystemDataConnector",
        module_name="great_expectations.datasource.data_connector.configured_asset_filesystem_data_connector",
        batch_spec_passthrough={
            "reader_options": {"header": True, "schema": spark_schema},
        },
    )


@pytest.fixture
def asset_config_spark(spark_session) -> AssetConfig:
    return AssetConfig(
        class_name="Asset",
        module_name="great_expectations.datasource.data_connector.asset",
        batch_spec_passthrough={
            "reader_options": {"header": True},
        },
    )


@pytest.fixture
def asset_config_with_schema_spark(spark_session, spark_schema) -> AssetConfig:
    return AssetConfig(
        class_name="Asset",
        module_name="great_expectations.datasource.data_connector.asset",
        batch_spec_passthrough={
            "reader_options": {"header": True, "schema": spark_schema},
        },
    )


@pytest.mark.unit
def test_lossy_serialization_warning(caplog):
    caplog.set_level(logging.WARNING, logger="great_expectations.core")

    d = Decimal("12345.678901234567890123456789")

    convert_to_json_serializable(d)
    assert len(caplog.messages) == 1
    assert caplog.messages[0].startswith(
        "Using lossy conversion for decimal 12345.678901234567890123456789"
    )

    caplog.clear()
    d = Decimal("0.1")
    convert_to_json_serializable(d)
    print(caplog.messages)
    assert len(caplog.messages) == 0


@pytest.mark.unit
def test_lossy_conversion():
    d = Decimal("12345.678901234567890123456789")
    assert requires_lossy_conversion(d)

    d = Decimal("12345.67890123456")
    assert requires_lossy_conversion(d)

    d = Decimal("12345.6789012345")
    assert not requires_lossy_conversion(d)

    d = Decimal("0.12345678901234567890123456789")
    assert requires_lossy_conversion(d)

    d = Decimal("0.1234567890123456")
    assert requires_lossy_conversion(d)

    d = Decimal("0.123456789012345")
    assert not requires_lossy_conversion(d)

    d = Decimal("0.1")
    assert not requires_lossy_conversion(d)


# TODO add unittests for convert_to_json_serializable() and ensure_json_serializable()
@pytest.mark.spark
def test_serialization_of_spark_df(spark_session):
    df = pd.DataFrame({"a": [1, 2, 3]})
    sdf = spark_session.createDataFrame(df)
    assert convert_to_json_serializable(sdf) == {"a": [1, 2, 3]}

    df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    sdf = spark_session.createDataFrame(df)
    assert convert_to_json_serializable(sdf) == {"a": [1, 2, 3], "b": [4, 5, 6]}


@pytest.mark.unit
def test_batch_request_deepcopy():
    test_df: pd.DataFrame = pd.DataFrame(data={"col1": [1, 2], "col2": [3, 4]})
    batch_request: RuntimeBatchRequest = RuntimeBatchRequest(
        **{
            "datasource_name": "my_datasource",
            "data_connector_name": "my_runtime_data_connector",
            "data_asset_name": "default_data_asset_name",
            "batch_identifiers": {
                "pipeline_stage_name": "core_processing",
                "airflow_run_id": 1234567890,
            },
            "runtime_parameters": {"batch_data": test_df},
        }
    )

    batch_request_copy: RuntimeBatchRequest = copy.deepcopy(batch_request)
    assert deep_filter_properties_iterable(
        properties=batch_request_copy.to_dict(),
        clean_falsy=True,
    ) == deep_filter_properties_iterable(
        properties=batch_request.to_dict(),
        clean_falsy=True,
    )


@pytest.mark.parametrize(
    "data_connector_config,expected_serialized_data_connector_config",
    [
        pytest.param(
            "data_connector_config_spark",
            {
                "batch_spec_passthrough": {
                    "reader_options": {
                        "header": True,
                    }
                },
                "class_name": "ConfiguredAssetFilesystemDataConnector",
                "module_name": "great_expectations.datasource.data_connector.configured_asset_filesystem_data_connector",  # noqa: E501 # FIXME CoP
            },
            id="data_connector_with_schema",
        ),
        pytest.param(
            "datas_connector_config_with_schema_spark",
            {
                "batch_spec_passthrough": {
                    "reader_options": {
                        "header": True,
                        "schema": {
                            "fields": [
                                {
                                    "metadata": {},
                                    "name": "a",
                                    "nullable": True,
                                    "type": "integer",
                                },
                                {
                                    "metadata": {},
                                    "name": "b",
                                    "nullable": True,
                                    "type": "integer",
                                },
                            ],
                            "type": "struct",
                        },
                    }
                },
                "class_name": "ConfiguredAssetFilesystemDataConnector",
                "module_name": "great_expectations.datasource.data_connector.configured_asset_filesystem_data_connector",  # noqa: E501 # FIXME CoP
            },
            id="data_connector_without_schema",
        ),
    ],
)
@pytest.mark.spark
def test_data_connector_and_nested_objects_are_serialized_spark(
    data_connector_config: DataConnectorConfig,
    expected_serialized_data_connector_config: dict,
    spark_session: pyspark.SparkSession,
    request: FixtureRequest,
):
    # when using a fixture value in a parmeterized test, we need to call
    # request.getfixturevalue()
    if isinstance(data_connector_config, str):
        data_connector_config = request.getfixturevalue(data_connector_config)

    observed_dump = dataConnectorConfigSchema.dump(obj=data_connector_config)
    assert observed_dump == expected_serialized_data_connector_config
    observed_load = dataConnectorConfigSchema.load(observed_dump)
    assert dataConnectorConfigSchema.dump(observed_load) == dataConnectorConfigSchema.dump(
        data_connector_config
    )


@pytest.mark.parametrize(
    "asset_config,expected_serialized_asset_config",
    [
        pytest.param(
            "asset_config_with_schema_spark",
            {
                "batch_spec_passthrough": {
                    "reader_options": {
                        "header": True,
                        "schema": {
                            "fields": [
                                {
                                    "metadata": {},
                                    "name": "a",
                                    "nullable": True,
                                    "type": "integer",
                                },
                                {
                                    "metadata": {},
                                    "name": "b",
                                    "nullable": True,
                                    "type": "integer",
                                },
                            ],
                            "type": "struct",
                        },
                    }
                },
                "class_name": "Asset",
                "module_name": "great_expectations.datasource.data_connector.asset",
            },
            id="asset_serialized_with_schema",
        ),
        pytest.param(
            "asset_config_spark",
            {
                "batch_spec_passthrough": {
                    "reader_options": {"header": True},
                },
                "class_name": "Asset",
                "module_name": "great_expectations.datasource.data_connector.asset",
            },
            id="asset_serialized_without_schema",
        ),
    ],
)
@pytest.mark.spark
def test_asset_and_nested_objects_are_serialized_spark(
    asset_config: AssetConfig,
    expected_serialized_asset_config: dict,
    spark_session: pyspark.SparkSession,
    request: FixtureRequest,
):
    # when using a fixture value in a parmeterized test, we need to call
    # request.getfixturevalue()
    if isinstance(asset_config, str):
        asset_config = request.getfixturevalue(asset_config)

    observed_dump = assetConfigSchema.dump(obj=asset_config)
    assert observed_dump == expected_serialized_asset_config
    observed_load = assetConfigSchema.load(observed_dump)
    assert assetConfigSchema.dump(observed_load) == assetConfigSchema.dump(asset_config)


def generic_config_serialization_assertions(
    config: AbstractConfig,
    schema: Schema,
    expected_serialized_config: dict,
    expected_roundtrip_config: dict,
) -> None:
    """Generic assertions for testing configuration serialization.

    Args:
        config: config object to check serialization.
        schema: Marshmallow schema used for serializing / deserializing config object.
        expected_serialized_config: expected when serializing a config via dump
        expected_roundtrip_config: expected config after loading the dump

    Returns:
        None

    Raises:
        AssertionError
    """

    observed_dump = schema.dump(config)
    assert observed_dump == expected_serialized_config

    loaded_data = schema.load(observed_dump)
    assert loaded_data.to_json_dict() == expected_roundtrip_config



import pytest

from great_expectations.render.view import DefaultJinjaPageView

# module level markers
pytestmark = pytest.mark.unit


def test_render_template():
    assert DefaultJinjaPageView().render_string_template(
        {
            "template": "It was the $first_adj of times; it was the $second_adj of times.",
            "params": {
                "first_adj": "best",
                "second_adj": "worst",
            },
        }
    ).replace(" ", "").replace("\t", "").replace(
        "\n", ""
    ) == "<span>It was the best of times; it was the worst of times.</span>".replace(
        " ", ""
    ).replace("\t", "").replace("\n", "")

    assert (
        DefaultJinjaPageView()
        .render_string_template(
            {
                "template": "It was the $first_adj of times; it was the $second_adj of times.",
                "params": {
                    "first_adj": "best",
                    "second_adj": "worst",
                },
                "styling": {
                    "default": {
                        "classes": ["badge", "badge-warning"],
                    }
                },
            }
        )
        .replace(" ", "")
        .replace("\t", "")
        .replace("\n", "")
        == '<span>It was the <span class="badge badge-warning" >best</span> of times; it was the <span class="badge badge-warning" >worst</span> of times.</span>'.replace(  # noqa: E501 # FIXME CoP
            " ", ""
        )
        .replace("\t", "")
        .replace("\n", "")
    )

    assert (
        DefaultJinjaPageView()
        .render_string_template(
            {
                "template": "It was the $first_adj of times; it was the $second_adj of times.",
                "params": {
                    "first_adj": "best",
                    "second_adj": "worst",
                },
                "styling": {
                    "default": {
                        "classes": ["badge", "badge-warning"],
                    },
                    "params": {
                        "first_adj": {
                            "classes": ["badge-error"],
                        }
                    },
                },
            }
        )
        .replace(" ", "")
        .replace("\t", "")
        .replace("\n", "")
        == '<span>It was the <span class="badge-error" >best</span> of times; it was the <span class="badge badge-warning" >worst</span> of times.</span>'.replace(  # noqa: E501 # FIXME CoP
            " ", ""
        )
        .replace("\t", "")
        .replace("\n", "")
    )

    assert (
        DefaultJinjaPageView()
        .render_string_template(
            {
                "template": "It was the $first_adj of times; it was the $second_adj of times.",
                "params": {
                    "first_adj": "best",
                    "second_adj": "worst",
                },
                "styling": {
                    "params": {
                        "first_adj": {
                            "classes": ["badge", "badge-warning"],
                        }
                    }
                },
            }
        )
        .replace(" ", "")
        .replace("\t", "")
        .replace("\n", "")
        == '<span>It was the <span class="badge badge-warning" >best</span> of times; it was the worst of times.</span>'.replace(  # noqa: E501 # FIXME CoP
            " ", ""
        )
        .replace("\t", "")
        .replace("\n", "")
    )

    assert (
        DefaultJinjaPageView()
        .render_string_template(
            {
                "template": "It was the $first_adj of times; it was the $second_adj of times.",
                "params": {
                    "first_adj": "best",
                    "second_adj": "worst",
                },
                "styling": {
                    "params": {
                        "first_adj": {
                            "classes": ["badge", "badge-warning"],
                            "attributes": {"role": "alert"},
                            "styles": {"padding": "5px"},
                        }
                    }
                },
            }
        )
        .replace(" ", "")
        .replace("\t", "")
        .replace("\n", "")
        == '<span>It was the <span class="badge badge-warning" role="alert" style="padding:5px;" >best</span> of times; it was the worst of times.</span>'.replace(  # noqa: E501 # FIXME CoP
            " ", ""
        )
        .replace("\t", "")
        .replace("\n", "")
    )


def test_render_template_with_extra_dollar_signs_in_template():
    result = DefaultJinjaPageView().render_string_template(
        {
            "template": "It was the $first_adj of times; it was the $second_adj of times. Blahhh$hhhh. $Bloooop. "  # noqa: E501 # FIXME CoP
            "Bleep$.",
            "params": {
                "first_adj": "best",
                "second_adj": "worst",
            },
        }
    )
    assert (
        result.replace(" ", "").replace("\t", "").replace("\n", "")
        == "<span>It was the best of times; it was the worst of times. Blahhh$hhhh. $Bloooop. Bleep$.</span>".replace(  # noqa: E501 # FIXME CoP
            " ", ""
        )
        .replace("\t", "")
        .replace("\n", "")
    )

    result = DefaultJinjaPageView().render_string_template(
        {
            "template": "It was the $first_adj of times; it was the $second_adj of times. Blahhh$$$hhhh. $$Bloooop. Bleep$$$$$.",  # noqa: E501 # FIXME CoP
            "params": {
                "first_adj": "best",
                "second_adj": "worst",
            },
        }
    )
    assert (
        result.replace(" ", "").replace("\t", "").replace("\n", "")
        == "<span>It was the best of times; it was the worst of times. Blahhh$$$hhhh. $$Bloooop. Bleep$$$$$.</span>".replace(  # noqa: E501 # FIXME CoP
            " ", ""
        )
        .replace("\t", "")
        .replace("\n", "")
    )


def test_render_template_with_extra_dollar_signs_in_param_values():
    result = DefaultJinjaPageView().render_string_template(
        {
            "template": "It was the $first_adj of times; it was the $second_adj of times. Blahhh$hhhh. $Bloooop. "  # noqa: E501 # FIXME CoP
            "Bleep$.",
            "params": {
                "first_adj": "$best$",
                "second_adj": "$$worst$",
            },
        }
    )
    assert (
        result.replace(" ", "").replace("\t", "").replace("\n", "")
        == "<span>It was the $best$ of times; it was the $$worst$ of times. Blahhh$hhhh. $Bloooop. Bleep$.</span>".replace(  # noqa: E501 # FIXME CoP
            " ", ""
        )
        .replace("\t", "")
        .replace("\n", "")
    )



from datetime import datetime
from unittest.mock import ANY

import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    JUST_PANDAS_DATA_SOURCES,
    NON_SQL_DATA_SOURCES,
    SQL_DATA_SOURCES,
)

NUMBERS_COLUMN = "numbers"
STRINGS_COLUMN = "strings"
DATES_COLUMN = "dates"
NULLS_COLUMN = "nulls"

DATA = pd.DataFrame(
    {
        NUMBERS_COLUMN: [1, 2, 3],
        STRINGS_COLUMN: ["a", "b", "c"],
        DATES_COLUMN: [
            datetime(2024, 1, 1).date(),  # noqa: DTZ001 # FIXME CoP
            datetime(2024, 2, 1).date(),  # noqa: DTZ001 # FIXME CoP
            datetime(2024, 3, 1).date(),  # noqa: DTZ001 # FIXME CoP
        ],
        NULLS_COLUMN: [1, None, 3],
    },
    dtype="object",
)


@parameterize_batch_for_data_sources(data_source_configs=NON_SQL_DATA_SOURCES, data=DATA)
def test_success_complete_non_sql(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValuesToBeInSet(column=NUMBERS_COLUMN, value_set=[1, 2, 3])
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert result.success


@parameterize_batch_for_data_sources(data_source_configs=SQL_DATA_SOURCES, data=DATA)
def test_success_complete_sql(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValuesToBeInSet(column=NUMBERS_COLUMN, value_set=[1, 2, 3])
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert result.success
    assert result.to_json_dict()["result"] == {
        "element_count": 3,
        "unexpected_count": 0,
        "unexpected_percent": 0.0,
        "partial_unexpected_list": [],
        "missing_count": 0,
        "missing_percent": 0.0,
        "unexpected_percent_total": 0.0,
        "unexpected_percent_nonmissing": 0.0,
        "partial_unexpected_counts": [],
        "unexpected_list": [],
        "unexpected_index_query": ANY,
    }


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnValuesToBeInSet(column=NUMBERS_COLUMN, value_set=[1, 2, 3]),
            id="basic_number_set",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInSet(column=STRINGS_COLUMN, value_set=["a", "b", "c"]),
            id="string_set",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInSet(
                column=DATES_COLUMN,
                value_set=[
                    datetime(2024, 1, 1).date(),  # noqa: DTZ001 # FIXME CoP
                    datetime(2024, 2, 1).date(),  # noqa: DTZ001 # FIXME CoP
                    datetime(2024, 3, 1).date(),  # noqa: DTZ001 # FIXME CoP
                ],
            ),
            id="date_set",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInSet(column=NUMBERS_COLUMN, value_set=[1, 2, 3, 4]),
            id="superset",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInSet(column=NUMBERS_COLUMN, value_set=[1, 2], mostly=0.66),
            id="mostly_in_set",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInSet(column=NULLS_COLUMN, value_set=[1, 3]),
            id="ignore_nulls",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success(
    batch_for_datasource: Batch,
    expectation: gxe.ExpectColumnValuesToBeInSet,
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnValuesToBeInSet(column=NUMBERS_COLUMN, value_set=[1, 2]),
            id="incomplete_number_set",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInSet(column=NUMBERS_COLUMN, value_set=[1, 2], mostly=0.7),
            id="mostly_threshold_not_met",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInSet(column=NUMBERS_COLUMN, value_set=[]),
            id="empty_set",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_failure(
    batch_for_datasource: Batch,
    expectation: gxe.ExpectColumnValuesToBeInSet,
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert not result.success


@parameterize_batch_for_data_sources(
    data_source_configs=JUST_PANDAS_DATA_SOURCES, data=pd.DataFrame({NUMBERS_COLUMN: []})
)
def test_empty_data_empty_set(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValuesToBeInSet(column=NUMBERS_COLUMN, value_set=[])
    result = batch_for_datasource.validate(expectation)
    assert result.success



from __future__ import annotations

import abc
import uuid
from typing import TYPE_CHECKING, Generic, TypeVar

if TYPE_CHECKING:
    from great_expectations.data_context import CloudDataContext

T = TypeVar("T")


class DataStore(abc.ABC, Generic[T]):
    """Abstract base class for all DataStore implementations."""

    def __init__(self, context: CloudDataContext):
        self._context = context

    @abc.abstractmethod
    def add(self, value: T) -> uuid.UUID:
        """Add a value to the DataStore.

        Args:
            value: Value to add to the DataStore.

        Returns:
            id of the created resource.
        """
        raise NotImplementedError



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import pytest

from great_expectations.data_context.util import file_relative_path
from great_expectations.expectations import UnexpectedRowsExpectation
from great_expectations.expectations.metrics.util import MAX_RESULT_RECORDS
from great_expectations.render.renderer.content_block.content_block import ContentBlockRenderer

if TYPE_CHECKING:
    from great_expectations.data_context import AbstractDataContext
    from great_expectations.datasource.fluent.interfaces import Batch
    from great_expectations.datasource.fluent.sqlite_datasource import SqliteDatasource


@pytest.fixture
def taxi_db_path() -> str:
    return file_relative_path(__file__, "../../test_sets/quickstart/yellow_tripdata.db")


@pytest.fixture
def sqlite_datasource(
    in_memory_runtime_context: AbstractDataContext, taxi_db_path: str
) -> SqliteDatasource:
    context = in_memory_runtime_context
    datasource_name = "my_sqlite_datasource"
    return context.data_sources.add_sqlite(
        datasource_name, connection_string=f"sqlite:///{taxi_db_path}"
    )


@pytest.fixture
def sqlite_batch(sqlite_datasource: SqliteDatasource) -> Batch:
    datasource = sqlite_datasource
    asset = datasource.add_table_asset("yellow_tripdata_sample_2022_01")

    batch_request = asset.build_batch_request()
    return asset.get_batch(batch_request)


@pytest.mark.unit
@pytest.mark.parametrize(
    "query",
    [
        pytest.param("SELECT * FROM table", id="no batch"),
        pytest.param("SELECT * FROM {{ batch }}", id="invalid format"),
        pytest.param("SELECT * FROM {active_batch}", id="legacy syntax"),
    ],
)
def test_unexpected_rows_expectation_invalid_query_info_message(query: str, caplog, capfd):
    # info log is emitted
    with caplog.at_level(logging.INFO):
        UnexpectedRowsExpectation(unexpected_rows_query=query)

    # stdout is printed to console
    out, _ = capfd.readouterr()
    assert "{batch}" in out


@pytest.mark.sqlite
@pytest.mark.parametrize(
    "query, expected_success, expected_observed_value, expected_count_unexpected_rows_returned",
    [
        pytest.param(
            "SELECT * FROM {batch} WHERE passenger_count > 7",
            True,
            0,
            0,
            id="success",
        ),
        pytest.param(
            # There is a single instance where passenger_count == 7
            "SELECT * FROM {batch} WHERE passenger_count > 6",
            False,
            1,
            1,
            id="failure",
        ),
        pytest.param(
            "SELECT * FROM {batch} WHERE passenger_count > 0",
            False,
            97853,
            MAX_RESULT_RECORDS,
            id="greater than MAX_RESULT_RECORDS unexpected rows",
        ),
    ],
)
def test_unexpected_rows_expectation_validate(
    sqlite_batch: Batch,
    query: str,
    expected_success: bool,
    expected_observed_value: int,
    expected_count_unexpected_rows_returned: int,
):
    batch = sqlite_batch

    expectation = UnexpectedRowsExpectation(unexpected_rows_query=query)
    result = batch.validate(expectation)

    assert result.success is expected_success

    res = result.result
    assert res["observed_value"] == expected_observed_value

    unexpected_count_rows_returned = len(res["details"]["unexpected_rows"])
    assert unexpected_count_rows_returned == expected_count_unexpected_rows_returned


@pytest.mark.unit
def test_unexpected_rows_expectation_correctly_interprets_query(
    sqlite_batch: Batch,
):
    query = "SELECT * FROM {batch}\r\n\t  ;\v\r ;"

    expectation = UnexpectedRowsExpectation(unexpected_rows_query=query)

    assert expectation.unexpected_rows_query == "SELECT * FROM {batch}"


@pytest.mark.unit
@pytest.mark.parametrize(
    "description, unexpected_rows_query",
    [
        pytest.param(
            "passenger_count should be less than or equal to 7",
            "SELECT * FROM {batch} WHERE passenger_count > 7",
            id="with description",
        ),
        pytest.param(
            None,
            "SELECT * FROM {batch} WHERE passenger_count > 7",
            id="no description",
        ),
    ],
)
def test_unexpected_rows_expectation_render(
    description: str | None,
    unexpected_rows_query: str,
):
    expectation = UnexpectedRowsExpectation(
        description=description,
        unexpected_rows_query=unexpected_rows_query,
    )
    expectation.render()
    assert (
        expectation.rendered_content[0].value.params.get("unexpected_rows_query").get("value")
        == unexpected_rows_query
    )

    assert expectation.rendered_content[0].value.template == description
    assert (
        expectation.rendered_content[0].value.code_block.get("code_template_str")
        == "$unexpected_rows_query"
    )
    assert expectation.rendered_content[0].value.code_block.get("language") == "sql"


@pytest.mark.unit
def test_data_docs_rendering():
    query = "SELECT * FROM {batch} WHERE passenger_count > 7"
    expectation = UnexpectedRowsExpectation(unexpected_rows_query=query)
    results = ContentBlockRenderer.render(expectation.configuration)
    assert isinstance(results, list) and len(results) == 1
    result = results[0]
    assert result.string_template == {
        "template": "Unexpected rows query: $unexpected_rows_query",
        "params": {"unexpected_rows_query": query},
        "styling": {},
    }



import pathlib
import re
from typing import TYPE_CHECKING, List, Union

import pytest

from great_expectations.compatibility import pydantic
from great_expectations.core import IDDict
from great_expectations.core.batch import LegacyBatchDefinition
from great_expectations.core.partitioners import FileNamePartitionerPath, FileNamePartitionerYearly
from great_expectations.datasource.fluent import BatchRequest
from great_expectations.datasource.fluent.constants import MATCH_ALL_PATTERN
from great_expectations.datasource.fluent.data_connector import (
    FilesystemDataConnector,
)
from tests.test_utils import create_files_in_directory

if TYPE_CHECKING:
    from great_expectations.datasource.fluent.data_connector import (
        DataConnector,
    )


@pytest.mark.filesystem
@pytest.mark.slow  # creating small number of`file handles in temporary file system
def test_basic_instantiation(tmp_path_factory):
    base_directory = str(tmp_path_factory.mktemp("test_basic_instantiation"))
    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "alpha-1.csv",
            "alpha-2.csv",
            "alpha-3.csv",
        ],
    )

    my_data_connector: DataConnector = FilesystemDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory),
        glob_directive="*.csv",
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "alpha-1.csv",
        "alpha-2.csv",
        "alpha-3.csv",
    ]
    assert my_data_connector.get_matched_data_reference_count() == 3
    assert my_data_connector.get_matched_data_references()[:3] == [
        "alpha-1.csv",
        "alpha-2.csv",
        "alpha-3.csv",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert my_data_connector.get_unmatched_data_reference_count() == 0

    # Missing "data_asset_name" argument.
    with pytest.raises(pydantic.ValidationError):
        # noinspection PyArgumentList
        my_data_connector.get_batch_definition_list(
            BatchRequest(
                datasource_name="something",
                options={},
            )
        )


@pytest.mark.filesystem
@pytest.mark.slow  # creating small number of`file handles in temporary file system
def test_return_all_batch_definitions_unsorted(tmp_path_factory):
    base_directory = str(tmp_path_factory.mktemp("test_return_all_batch_definitions_unsorted"))
    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "alex_20200809_1000.csv",
            "eugene_20200809_1500.csv",
            "james_20200811_1009.csv",
            "abe_20200809_1040.csv",
            "will_20200809_1002.csv",
            "james_20200713_1567.csv",
            "eugene_20201129_1900.csv",
            "will_20200810_1001.csv",
            "james_20200810_1003.csv",
            "alex_20200819_1300.csv",
        ],
    )
    batching_regex = re.compile(r"(?P<name>.+)_(?P<timestamp>.+)_(?P<price>.+)\.csv")
    my_data_connector: DataConnector = FilesystemDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory),
        glob_directive="*.csv",
    )
    # with missing BatchRequest arguments
    with pytest.raises(TypeError):
        # noinspection PyArgumentList
        my_data_connector.get_batch_definition_list()

    # with empty options
    unsorted_batch_definition_list: List[LegacyBatchDefinition] = (
        my_data_connector.get_batch_definition_list(
            BatchRequest(
                datasource_name="my_file_path_datasource",
                data_asset_name="my_filesystem_data_asset",
                options={},
                partitioner=FileNamePartitionerPath(regex=batching_regex),
            )
        )
    )
    processed_batching_regex = re.compile(
        "(?P<path>(?P<name>.+)_(?P<timestamp>.+)_(?P<price>.+)\\.csv)"
    )
    expected: List[LegacyBatchDefinition] = [
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "abe_20200809_1040.csv",
                    "name": "abe",
                    "timestamp": "20200809",
                    "price": "1040",
                }
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "alex_20200809_1000.csv",
                    "name": "alex",
                    "timestamp": "20200809",
                    "price": "1000",
                }
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "alex_20200819_1300.csv",
                    "name": "alex",
                    "timestamp": "20200819",
                    "price": "1300",
                }
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "eugene_20200809_1500.csv",
                    "name": "eugene",
                    "timestamp": "20200809",
                    "price": "1500",
                }
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "eugene_20201129_1900.csv",
                    "name": "eugene",
                    "timestamp": "20201129",
                    "price": "1900",
                }
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "james_20200713_1567.csv",
                    "name": "james",
                    "timestamp": "20200713",
                    "price": "1567",
                }
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "james_20200810_1003.csv",
                    "name": "james",
                    "timestamp": "20200810",
                    "price": "1003",
                }
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "james_20200811_1009.csv",
                    "name": "james",
                    "timestamp": "20200811",
                    "price": "1009",
                }
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "will_20200809_1002.csv",
                    "name": "will",
                    "timestamp": "20200809",
                    "price": "1002",
                }
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "will_20200810_1001.csv",
                    "name": "will",
                    "timestamp": "20200810",
                    "price": "1001",
                }
            ),
            batching_regex=processed_batching_regex,
        ),
    ]
    assert expected == unsorted_batch_definition_list

    # with specified Batch query options
    unsorted_batch_definition_list = my_data_connector.get_batch_definition_list(
        BatchRequest(
            datasource_name="my_file_path_datasource",
            data_asset_name="my_filesystem_data_asset",
            options={"name": "alex", "timestamp": "20200819", "price": "1300"},
            partitioner=FileNamePartitionerPath(regex=batching_regex),
        )
    )
    assert expected[2:3] == unsorted_batch_definition_list


@pytest.mark.filesystem
@pytest.mark.slow  # creating small number of`file handles in temporary file system
def test_return_only_unique_batch_definitions(tmp_path_factory):
    base_directory = str(tmp_path_factory.mktemp("test_return_only_unique_batch_definitions"))
    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "A/file_1.csv",
            "A/file_2.csv",
            "A/file_3.csv",
            "B/file_1.csv",
            "B/file_2.csv",
        ],
    )

    my_data_connector: DataConnector

    my_data_connector = FilesystemDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory),
    )

    assert my_data_connector.get_data_reference_count() == 7
    assert my_data_connector.get_data_references()[:3] == [
        "A",
        "A/file_1.csv",
        "A/file_2.csv",
    ]

    batching_regex = re.compile(r"(?P<name>.+)/.+\.csv")
    matched_data_references = my_data_connector.get_matched_data_references(regex=batching_regex)
    assert len(matched_data_references) == 5
    assert matched_data_references[:3] == [
        "A/file_1.csv",
        "A/file_2.csv",
        "A/file_3.csv",
    ]

    processed_batching_regex = re.compile("(?P<path>(?P<directory>.+)/(?P<filename>.+\\.csv))")
    expected: List[LegacyBatchDefinition] = [
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {"path": "A/file_1.csv", "directory": "A", "filename": "file_1.csv"}
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {"path": "A/file_2.csv", "directory": "A", "filename": "file_2.csv"}
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {"path": "A/file_3.csv", "directory": "A", "filename": "file_3.csv"}
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {"path": "B/file_1.csv", "directory": "B", "filename": "file_1.csv"}
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {"path": "B/file_2.csv", "directory": "B", "filename": "file_2.csv"}
            ),
            batching_regex=processed_batching_regex,
        ),
    ]

    my_data_connector = FilesystemDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory),
    )

    batching_regex = re.compile(r"(?P<directory>.+)/(?P<filename>.+\.csv)")
    unsorted_batch_definition_list: List[LegacyBatchDefinition] = (
        my_data_connector.get_batch_definition_list(
            BatchRequest(
                datasource_name="my_file_path_datasource",
                data_asset_name="my_filesystem_data_asset",
                options={},
                partitioner=FileNamePartitionerPath(regex=batching_regex),
            )
        )
    )
    assert expected == unsorted_batch_definition_list


@pytest.mark.filesystem
@pytest.mark.slow  # creating small number of`file handles in temporary file system
def test_alpha(tmp_path_factory):
    base_directory = str(tmp_path_factory.mktemp("test_alpha"))
    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "test_dir_alpha/A.csv",
            "test_dir_alpha/B.csv",
            "test_dir_alpha/C.csv",
            "test_dir_alpha/D.csv",
        ],
    )

    my_data_connector: DataConnector = FilesystemDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory) / "test_dir_alpha",
        glob_directive="*.csv",
    )
    assert my_data_connector.get_data_reference_count() == 4
    assert my_data_connector.get_data_references()[:3] == [
        "A.csv",
        "B.csv",
        "C.csv",
    ]

    batching_regex = re.compile(r"(?P<part_1>.+)\.csv")
    matched_data_references = my_data_connector.get_matched_data_references(regex=batching_regex)
    assert len(matched_data_references) == 4
    assert matched_data_references[:3] == [
        "A.csv",
        "B.csv",
        "C.csv",
    ]

    # Try to fetch a batch from a nonexistent asset
    my_batch_request = BatchRequest(
        datasource_name="BASE",
        data_asset_name="A",
        options={},
        partitioner=FileNamePartitionerPath(regex=batching_regex),
    )
    my_batch_definition_list = my_data_connector.get_batch_definition_list(
        batch_request=my_batch_request
    )
    assert len(my_batch_definition_list) == 0

    my_batch_request = BatchRequest(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_filesystem_data_asset",
        options={"part_1": "B"},
        partitioner=FileNamePartitionerPath(regex=batching_regex),
    )
    my_batch_definition_list = my_data_connector.get_batch_definition_list(
        batch_request=my_batch_request
    )
    assert len(my_batch_definition_list) == 1


@pytest.mark.filesystem
@pytest.mark.slow  # creating small number of`file handles in temporary file system
def test_foxtrot(tmp_path_factory):
    base_directory = str(tmp_path_factory.mktemp("test_foxtrot"))
    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "test_dir_foxtrot/A/A-1.csv",
            "test_dir_foxtrot/A/A-2.csv",
            "test_dir_foxtrot/A/A-3.csv",
            "test_dir_foxtrot/B/B-1.txt",
            "test_dir_foxtrot/B/B-2.txt",
            "test_dir_foxtrot/B/B-3.txt",
            "test_dir_foxtrot/C/C-2017.csv",
            "test_dir_foxtrot/C/C-2018.csv",
            "test_dir_foxtrot/C/C-2019.csv",
            "test_dir_foxtrot/D/D-aaa.csv",
            "test_dir_foxtrot/D/D-bbb.csv",
            "test_dir_foxtrot/D/D-ccc.csv",
            "test_dir_foxtrot/D/D-ddd.csv",
            "test_dir_foxtrot/D/D-eee.csv",
        ],
    )

    my_data_connector: DataConnector

    my_data_connector = FilesystemDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory) / "test_dir_foxtrot",
        glob_directive="*.csv",
    )

    batching_regex = re.compile(r"(?P<part_1>.+)-(?P<part_2>.+)\.csv")
    matched_data_references = my_data_connector.get_matched_data_references(regex=batching_regex)
    assert my_data_connector.get_data_reference_count() == 0
    assert my_data_connector.get_data_references()[:3] == []
    assert len(matched_data_references) == 0

    my_data_connector = FilesystemDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory) / "test_dir_foxtrot" / "A",
        glob_directive="*.csv",
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "A-1.csv",
        "A-2.csv",
        "A-3.csv",
    ]

    batching_regex = re.compile(r"(?P<part_1>.+)-(?P<part_2>.+)\.csv")
    matched_data_references = my_data_connector.get_matched_data_references(regex=batching_regex)
    assert len(matched_data_references) == 3
    assert matched_data_references[:3] == [
        "A-1.csv",
        "A-2.csv",
        "A-3.csv",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert my_data_connector.get_unmatched_data_reference_count() == 0

    my_data_connector = FilesystemDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory) / "test_dir_foxtrot" / "B",
        glob_directive="*.*",
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "B-1.txt",
        "B-2.txt",
        "B-3.txt",
    ]

    batching_regex = re.compile(r"(?P<part_1>.+)-(?P<part_2>.+)\.txt")
    matched_data_references = my_data_connector.get_matched_data_references(regex=batching_regex)
    assert len(matched_data_references) == 3
    assert matched_data_references[:3] == [
        "B-1.txt",
        "B-2.txt",
        "B-3.txt",
    ]
    assert my_data_connector.get_data_reference_count() == 3

    my_data_connector = FilesystemDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory) / "test_dir_foxtrot" / "C",
        glob_directive="*",
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "C-2017.csv",
        "C-2018.csv",
        "C-2019.csv",
    ]

    batching_regex = re.compile(r"(?P<part_1>.+)-(?P<part_2>.+)\.csv")
    matched_data_references = my_data_connector.get_matched_data_references(regex=batching_regex)
    assert len(matched_data_references) == 3
    assert matched_data_references[:3] == [
        "C-2017.csv",
        "C-2018.csv",
        "C-2019.csv",
    ]

    my_batch_request = BatchRequest(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_filesystem_data_asset",
        options={},
        partitioner=FileNamePartitionerPath(regex=batching_regex),
    )
    my_batch_definition_list: List[LegacyBatchDefinition] = (
        my_data_connector.get_batch_definition_list(batch_request=my_batch_request)
    )
    assert len(my_batch_definition_list) == 3


@pytest.mark.filesystem
@pytest.mark.slow  # creating small number of`file handles in temporary file system
def test_relative_base_directory_path(tmp_path_factory):
    base_directory = str(tmp_path_factory.mktemp("test_relative_asset_base_directory_path"))
    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "test_dir_0/A/B/C/logfile_0.csv",
            "test_dir_0/A/B/C/bigfile_1.csv",
            "test_dir_0/A/filename2.csv",
            "test_dir_0/A/filename3.csv",
        ],
    )

    my_data_connector = FilesystemDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory) / "test_dir_0" / "A",
        glob_directive="*",
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "B",
        "filename2.csv",
        "filename3.csv",
    ]

    batching_regex = re.compile(r"(?P<part_1>.+)\.csv")
    matched_data_references = my_data_connector.get_matched_data_references(regex=batching_regex)
    assert len(matched_data_references) == 2
    assert matched_data_references[:3] == [
        "filename2.csv",
        "filename3.csv",
    ]

    my_data_connector = FilesystemDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory) / "test_dir_0" / "A" / "B" / "C",
        glob_directive="log*.csv",
    )
    assert my_data_connector.get_data_reference_count() == 1
    assert my_data_connector.get_data_references()[:3] == [
        "logfile_0.csv",
    ]

    batching_regex = re.compile(r"(?P<name>.+)_(?P<number>.+)\.csv")
    matched_data_references = my_data_connector.get_matched_data_references(regex=batching_regex)
    assert len(matched_data_references) == 1
    assert matched_data_references[:3] == [
        "logfile_0.csv",
    ]
    assert (
        my_data_connector._get_full_file_path(path="bigfile_1.csv")
        == f"{base_directory}/test_dir_0/A/B/C/bigfile_1.csv"
    )

    my_batch_request: BatchRequest = BatchRequest(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_filesystem_data_asset",
        options={},
        partitioner=FileNamePartitionerPath(regex=batching_regex),
    )
    my_batch_definition_list: List[LegacyBatchDefinition] = (
        my_data_connector.get_batch_definition_list(batch_request=my_batch_request)
    )
    assert len(my_batch_definition_list) == 1


@pytest.mark.filesystem
@pytest.mark.parametrize(
    "batching_regex,batch_definition_count",
    [
        pytest.param(MATCH_ALL_PATTERN, 6, id="Match All Pattern"),
        pytest.param(r"alpha-(.*)\.csv", 3, id="Match Alpha CSV with Group"),
        pytest.param(r"single-match\..*", 1, id="Match a single CSV"),
        pytest.param(r"mismatch", 0, id="Match nothing"),
    ],
)
def test_filesystem_data_connector_uses_batching_regex_from_batch_request(
    tmp_path_factory, batching_regex: Union[str, re.Pattern], batch_definition_count
):
    # arrange
    base_directory = str(tmp_path_factory.mktemp("test_basic_instantiation"))
    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "alpha-1.csv",
            "alpha-2.csv",
            "alpha-3.csv",
            "dont-match.csv",
            "this-either.csv",
            "single-match.csv",
        ],
    )

    my_data_connector: DataConnector = FilesystemDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory),
        glob_directive="*.csv",
    )

    # act
    batch_definitions = my_data_connector.get_batch_definition_list(
        BatchRequest(
            datasource_name="my_file_path_datasource",
            data_asset_name="my_filesystem_data_asset",
            options={},
            partitioner=FileNamePartitionerYearly(regex=re.compile(batching_regex)),
        )
    )

    # assert
    assert len(batch_definitions) == batch_definition_count
